from typing import List, Optional, Type

from graphql.error import GraphQLError
from graphql.language import parse
from graphql.type import GraphQLSchema
from graphql.utilities import build_schema
from graphql.validation import ValidationRule, SDLValidationRule
from graphql.validation.validate import validate, validate_sdl

__all__ = [
    "test_schema",
    "assert_validation_errors",
    "assert_sdl_validation_errors",
]

test_schema = build_schema(
    """
    interface Mammal {
      mother: Mammal
      father: Mammal
    }

    interface Pet {
      name(surname: Boolean): String
    }

    interface Canine implements Mammal {
      name(surname: Boolean): String
      mother: Canine
      father: Canine
    }

    enum DogCommand {
      SIT
      HEEL
      DOWN
    }

    type Dog implements Pet & Mammal & Canine {
      name(surname: Boolean): String
      nickname: String
      barkVolume: Int
      barks: Boolean
      doesKnowCommand(dogCommand: DogCommand): Boolean
      isHouseTrained(atOtherHomes: Boolean = true): Boolean
      isAtLocation(x: Int, y: Int): Boolean
      mother: Dog
      father: Dog
    }

    type Cat implements Pet {
      name(surname: Boolean): String
      nickname: String
      meows: Boolean
      meowsVolume: Int
      furColor: FurColor
    }

    union CatOrDog = Cat | Dog

    type Human {
      name(surname: Boolean): String
      pets: [Pet]
      relatives: [Human]
    }

    enum FurColor {
      BROWN
      BLACK
      TAN
      SPOTTED
      NO_FUR
      UNKNOWN
    }

    input ComplexInput {
      requiredField: Boolean!
      nonNullField: Boolean! = false
      intField: Int
      stringField: String
      booleanField: Boolean
      stringListField: [String]
    }

    type ComplicatedArgs {
      # TODO List
      # TODO Coercion
      # TODO NotNulls
      intArgField(intArg: Int): String
      nonNullIntArgField(nonNullIntArg: Int!): String
      stringArgField(stringArg: String): String
      booleanArgField(booleanArg: Boolean): String
      enumArgField(enumArg: FurColor): String
      floatArgField(floatArg: Float): String
      idArgField(idArg: ID): String
      stringListArgField(stringListArg: [String]): String
      stringListNonNullArgField(stringListNonNullArg: [String!]): String
      complexArgField(complexArg: ComplexInput): String
      multipleReqs(req1: Int!, req2: Int!): String
      nonNullFieldWithDefault(arg: Int! = 0): String
      multipleOpts(opt1: Int = 0, opt2: Int = 0): String
      multipleOptAndReq(req1: Int!, req2: Int!, opt1: Int = 0, opt2: Int = 0): String
    }

    type QueryRoot {
      human(id: ID): Human
      dog: Dog
      cat: Cat
      pet: Pet
      catOrDog: CatOrDog
      complicatedArgs: ComplicatedArgs
    }

    schema {
      query: QueryRoot
    }

    directive @onField on FIELD
    """
)


def assert_validation_errors(
    rule: Type[ValidationRule],
    query_str: str,
    errors: List[GraphQLError],
    schema: GraphQLSchema = test_schema,
) -> List[GraphQLError]:
    doc = parse(query_str)
    returned_errors = validate(schema, doc, [rule])
    assert returned_errors == errors
    return returned_errors


def assert_sdl_validation_errors(
    rule: Type[SDLValidationRule],
    sdl_str: str,
    errors: List[GraphQLError],
    schema: Optional[GraphQLSchema] = None,
) -> List[GraphQLError]:
    doc = parse(sdl_str)
    returned_errors = validate_sdl(doc, schema, [rule])
    assert returned_errors == errors
    return returned_errors



"""GraphQL-core

The primary :mod:`graphql` package includes everything you need to define a GraphQL
schema and fulfill GraphQL requests.

GraphQL-core provides a reference implementation for the GraphQL specification
but is also a useful utility for operating on GraphQL files and building sophisticated
tools.

This top-level package exports a general purpose function for fulfilling all steps
of the GraphQL specification in a single operation, but also includes utilities
for every part of the GraphQL specification:

  - Parsing the GraphQL language.
  - Building a GraphQL type schema.
  - Validating a GraphQL request against a type schema.
  - Executing a GraphQL request against a type schema.

This also includes utility functions for operating on GraphQL types and GraphQL
documents to facilitate building tools.

You may also import from each sub-package directly. For example, the following two
import statements are equivalent::

    from graphql import parse
    from graphql.language import parse

The sub-packages of GraphQL-core 3 are:

  - :mod:`graphql.language`: Parse and operate on the GraphQL language.
  - :mod:`graphql.type`: Define GraphQL types and schema.
  - :mod:`graphql.validation`: The Validation phase of fulfilling a GraphQL result.
  - :mod:`graphql.execution`: The Execution phase of fulfilling a GraphQL request.
  - :mod:`graphql.error`: Creating and formatting GraphQL errors.
  - :mod:`graphql.utilities`:
    Common useful computations upon the GraphQL language and type objects.
"""

# The GraphQL-core 3 and GraphQL.js version info.

from .version import version, version_info, version_js, version_info_js

# Utilities for compatibility with the Python language.
from .pyutils import Undefined, UndefinedType

# Create, format, and print GraphQL errors.
from .error import (
    GraphQLError,
    GraphQLErrorExtensions,
    GraphQLFormattedError,
    GraphQLSyntaxError,
    located_error,
)

# Parse and operate on GraphQL language source files.
from .language import (
    Source,
    get_location,
    # Print source location
    print_location,
    print_source_location,
    # Lex
    Lexer,
    TokenKind,
    # Parse
    parse,
    parse_value,
    parse_const_value,
    parse_type,
    # Print
    print_ast,
    # Visit
    visit,
    ParallelVisitor,
    Visitor,
    VisitorAction,
    VisitorKeyMap,
    BREAK,
    SKIP,
    REMOVE,
    IDLE,
    DirectiveLocation,
    # Predicates
    is_definition_node,
    is_executable_definition_node,
    is_selection_node,
    is_value_node,
    is_const_value_node,
    is_type_node,
    is_type_system_definition_node,
    is_type_definition_node,
    is_type_system_extension_node,
    is_type_extension_node,
    # Types
    SourceLocation,
    Location,
    Token,
    # AST nodes
    Node,
    # Each kind of AST node
    NameNode,
    DocumentNode,
    DefinitionNode,
    ExecutableDefinitionNode,
    OperationDefinitionNode,
    OperationType,
    VariableDefinitionNode,
    VariableNode,
    SelectionSetNode,
    SelectionNode,
    FieldNode,
    ArgumentNode,
    ConstArgumentNode,
    FragmentSpreadNode,
    InlineFragmentNode,
    FragmentDefinitionNode,
    ValueNode,
    ConstValueNode,
    IntValueNode,
    FloatValueNode,
    StringValueNode,
    BooleanValueNode,
    NullValueNode,
    EnumValueNode,
    ListValueNode,
    ConstListValueNode,
    ObjectValueNode,
    ConstObjectValueNode,
    ObjectFieldNode,
    ConstObjectFieldNode,
    DirectiveNode,
    ConstDirectiveNode,
    TypeNode,
    NamedTypeNode,
    ListTypeNode,
    NonNullTypeNode,
    TypeSystemDefinitionNode,
    SchemaDefinitionNode,
    OperationTypeDefinitionNode,
    TypeDefinitionNode,
    ScalarTypeDefinitionNode,
    ObjectTypeDefinitionNode,
    FieldDefinitionNode,
    InputValueDefinitionNode,
    InterfaceTypeDefinitionNode,
    UnionTypeDefinitionNode,
    EnumTypeDefinitionNode,
    EnumValueDefinitionNode,
    InputObjectTypeDefinitionNode,
    DirectiveDefinitionNode,
    TypeSystemExtensionNode,
    SchemaExtensionNode,
    TypeExtensionNode,
    ScalarTypeExtensionNode,
    ObjectTypeExtensionNode,
    InterfaceTypeExtensionNode,
    UnionTypeExtensionNode,
    EnumTypeExtensionNode,
    InputObjectTypeExtensionNode,
)

# Utilities for operating on GraphQL type schema and parsed sources.
from .utilities import (
    # Produce the GraphQL query recommended for a full schema introspection.
    # Accepts optional IntrospectionOptions.
    get_introspection_query,
    IntrospectionQuery,
    # Get the target Operation from a Document.
    get_operation_ast,
    # Get the Type for the target Operation AST.
    get_operation_root_type,
    # Convert a GraphQLSchema to an IntrospectionQuery.
    introspection_from_schema,
    # Build a GraphQLSchema from an introspection result.
    build_client_schema,
    # Build a GraphQLSchema from a parsed GraphQL Schema language AST.
    build_ast_schema,
    # Build a GraphQLSchema from a GraphQL schema language document.
    build_schema,
    # Extend an existing GraphQLSchema from a parsed GraphQL Schema language AST.
    extend_schema,
    # Sort a GraphQLSchema.
    lexicographic_sort_schema,
    # Print a GraphQLSchema to GraphQL Schema language.
    print_schema,
    # Print a GraphQLType to GraphQL Schema language.
    print_type,
    # Prints the built-in introspection schema in the Schema Language format.
    print_introspection_schema,
    # Create a GraphQLType from a GraphQL language AST.
    type_from_ast,
    # Convert a language AST to a dictionary.
    ast_to_dict,
    # Create a Python value from a GraphQL language AST with a Type.
    value_from_ast,
    # Create a Python value from a GraphQL language AST without a Type.
    value_from_ast_untyped,
    # Create a GraphQL language AST from a Python value.
    ast_from_value,
    # A helper to use within recursive-descent visitors which need to be aware of the
    # GraphQL type system.
    TypeInfo,
    TypeInfoVisitor,
    # Coerce a Python value to a GraphQL type, or produce errors.
    coerce_input_value,
    # Concatenates multiple ASTs together.
    concat_ast,
    # Separate an AST into an AST per Operation.
    separate_operations,
    # Strip characters that are not significant to the validity or execution
    # of a GraphQL document.
    strip_ignored_characters,
    # Comparators for types
    is_equal_type,
    is_type_sub_type_of,
    do_types_overlap,
    # Assert a string is a valid GraphQL name.
    assert_valid_name,
    # Determine if a string is a valid GraphQL name.
    is_valid_name_error,
    # Compare two GraphQLSchemas and detect breaking changes.
    BreakingChange,
    BreakingChangeType,
    DangerousChange,
    DangerousChangeType,
    find_breaking_changes,
    find_dangerous_changes,
)

# Create and operate on GraphQL type definitions and schema.
from .type import (
    # Definitions
    GraphQLSchema,
    GraphQLDirective,
    GraphQLScalarType,
    GraphQLObjectType,
    GraphQLInterfaceType,
    GraphQLUnionType,
    GraphQLEnumType,
    GraphQLInputObjectType,
    GraphQLList,
    GraphQLNonNull,
    # Standard GraphQL Scalars
    specified_scalar_types,
    GraphQLInt,
    GraphQLFloat,
    GraphQLString,
    GraphQLBoolean,
    GraphQLID,
    # Int boundaries constants
    GRAPHQL_MAX_INT,
    GRAPHQL_MIN_INT,
    # Built-in Directives defined by the Spec
    specified_directives,
    GraphQLIncludeDirective,
    GraphQLSkipDirective,
    GraphQLDeprecatedDirective,
    GraphQLSpecifiedByDirective,
    # "Enum" of Type Kinds
    TypeKind,
    # Constant Deprecation Reason
    DEFAULT_DEPRECATION_REASON,
    # GraphQL Types for introspection.
    introspection_types,
    # Meta-field definitions.
    SchemaMetaFieldDef,
    TypeMetaFieldDef,
    TypeNameMetaFieldDef,
    # Predicates
    is_schema,
    is_directive,
    is_type,
    is_scalar_type,
    is_object_type,
    is_interface_type,
    is_union_type,
    is_enum_type,
    is_input_object_type,
    is_list_type,
    is_non_null_type,
    is_input_type,
    is_output_type,
    is_leaf_type,
    is_composite_type,
    is_abstract_type,
    is_wrapping_type,
    is_nullable_type,
    is_named_type,
    is_required_argument,
    is_required_input_field,
    is_specified_scalar_type,
    is_introspection_type,
    is_specified_directive,
    # Assertions
    assert_schema,
    assert_directive,
    assert_type,
    assert_scalar_type,
    assert_object_type,
    assert_interface_type,
    assert_union_type,
    assert_enum_type,
    assert_input_object_type,
    assert_list_type,
    assert_non_null_type,
    assert_input_type,
    assert_output_type,
    assert_leaf_type,
    assert_composite_type,
    assert_abstract_type,
    assert_wrapping_type,
    assert_nullable_type,
    assert_named_type,
    # Un-modifiers
    get_nullable_type,
    get_named_type,
    # Thunk handling
    resolve_thunk,
    # Validate GraphQL schema.
    validate_schema,
    assert_valid_schema,
    #  Uphold the spec rules about naming
    assert_name,
    assert_enum_value_name,
    # Types
    GraphQLType,
    GraphQLInputType,
    GraphQLOutputType,
    GraphQLLeafType,
    GraphQLCompositeType,
    GraphQLAbstractType,
    GraphQLWrappingType,
    GraphQLNullableType,
    GraphQLNamedType,
    GraphQLNamedInputType,
    GraphQLNamedOutputType,
    Thunk,
    ThunkCollection,
    ThunkMapping,
    GraphQLArgument,
    GraphQLArgumentMap,
    GraphQLEnumValue,
    GraphQLEnumValueMap,
    GraphQLField,
    GraphQLFieldMap,
    GraphQLFieldResolver,
    GraphQLInputField,
    GraphQLInputFieldMap,
    GraphQLScalarSerializer,
    GraphQLScalarValueParser,
    GraphQLScalarLiteralParser,
    GraphQLIsTypeOfFn,
    GraphQLResolveInfo,
    ResponsePath,
    GraphQLTypeResolver,
    # Keyword args
    GraphQLArgumentKwargs,
    GraphQLDirectiveKwargs,
    GraphQLEnumTypeKwargs,
    GraphQLEnumValueKwargs,
    GraphQLFieldKwargs,
    GraphQLInputFieldKwargs,
    GraphQLInputObjectTypeKwargs,
    GraphQLInterfaceTypeKwargs,
    GraphQLNamedTypeKwargs,
    GraphQLObjectTypeKwargs,
    GraphQLScalarTypeKwargs,
    GraphQLSchemaKwargs,
    GraphQLUnionTypeKwargs,
)

# Validate GraphQL queries.
from .validation import (
    validate,
    ValidationContext,
    ValidationRule,
    ASTValidationRule,
    SDLValidationRule,
    # All validation rules in the GraphQL Specification.
    specified_rules,
    # Individual validation rules.
    ExecutableDefinitionsRule,
    FieldsOnCorrectTypeRule,
    FragmentsOnCompositeTypesRule,
    KnownArgumentNamesRule,
    KnownDirectivesRule,
    KnownFragmentNamesRule,
    KnownTypeNamesRule,
    LoneAnonymousOperationRule,
    NoFragmentCyclesRule,
    NoUndefinedVariablesRule,
    NoUnusedFragmentsRule,
    NoUnusedVariablesRule,
    OverlappingFieldsCanBeMergedRule,
    PossibleFragmentSpreadsRule,
    ProvidedRequiredArgumentsRule,
    ScalarLeafsRule,
    SingleFieldSubscriptionsRule,
    UniqueArgumentNamesRule,
    UniqueDirectivesPerLocationRule,
    UniqueFragmentNamesRule,
    UniqueInputFieldNamesRule,
    UniqueOperationNamesRule,
    UniqueVariableNamesRule,
    ValuesOfCorrectTypeRule,
    VariablesAreInputTypesRule,
    VariablesInAllowedPositionRule,
    # SDL-specific validation rules
    LoneSchemaDefinitionRule,
    UniqueOperationTypesRule,
    UniqueTypeNamesRule,
    UniqueEnumValueNamesRule,
    UniqueFieldDefinitionNamesRule,
    UniqueArgumentDefinitionNamesRule,
    UniqueDirectiveNamesRule,
    PossibleTypeExtensionsRule,
    # Custom validation rules
    NoDeprecatedCustomRule,
    NoSchemaIntrospectionCustomRule,
)

# Execute GraphQL documents.
from .execution import (
    execute,
    execute_sync,
    default_field_resolver,
    default_type_resolver,
    get_argument_values,
    get_directive_values,
    get_variable_values,
    # Types
    ExecutionContext,
    ExecutionResult,
    FormattedExecutionResult,
    # Subscription
    subscribe,
    create_source_event_stream,
    MapAsyncIterator,
    # Middleware
    Middleware,
    MiddlewareManager,
)

# The primary entry point into fulfilling a GraphQL request.
from .graphql import graphql, graphql_sync

INVALID = Undefined  # deprecated alias

# The GraphQL-core version info.
__version__ = version
__version_info__ = version_info

# The GraphQL.js version info.
__version_js__ = version_js
__version_info_js__ = version_info_js


__all__ = [
    "version",
    "version_info",
    "version_js",
    "version_info_js",
    "graphql",
    "graphql_sync",
    "GraphQLSchema",
    "GraphQLDirective",
    "GraphQLScalarType",
    "GraphQLObjectType",
    "GraphQLInterfaceType",
    "GraphQLUnionType",
    "GraphQLEnumType",
    "GraphQLInputObjectType",
    "GraphQLList",
    "GraphQLNonNull",
    "specified_scalar_types",
    "GraphQLInt",
    "GraphQLFloat",
    "GraphQLString",
    "GraphQLBoolean",
    "GraphQLID",
    "GRAPHQL_MAX_INT",
    "GRAPHQL_MIN_INT",
    "specified_directives",
    "GraphQLIncludeDirective",
    "GraphQLSkipDirective",
    "GraphQLDeprecatedDirective",
    "GraphQLSpecifiedByDirective",
    "TypeKind",
    "DEFAULT_DEPRECATION_REASON",
    "introspection_types",
    "SchemaMetaFieldDef",
    "TypeMetaFieldDef",
    "TypeNameMetaFieldDef",
    "is_schema",
    "is_directive",
    "is_type",
    "is_scalar_type",
    "is_object_type",
    "is_interface_type",
    "is_union_type",
    "is_enum_type",
    "is_input_object_type",
    "is_list_type",
    "is_non_null_type",
    "is_input_type",
    "is_output_type",
    "is_leaf_type",
    "is_composite_type",
    "is_abstract_type",
    "is_wrapping_type",
    "is_nullable_type",
    "is_named_type",
    "is_required_argument",
    "is_required_input_field",
    "is_specified_scalar_type",
    "is_introspection_type",
    "is_specified_directive",
    "assert_schema",
    "assert_directive",
    "assert_type",
    "assert_scalar_type",
    "assert_object_type",
    "assert_interface_type",
    "assert_union_type",
    "assert_enum_type",
    "assert_input_object_type",
    "assert_list_type",
    "assert_non_null_type",
    "assert_input_type",
    "assert_output_type",
    "assert_leaf_type",
    "assert_composite_type",
    "assert_abstract_type",
    "assert_wrapping_type",
    "assert_nullable_type",
    "assert_named_type",
    "get_nullable_type",
    "get_named_type",
    "resolve_thunk",
    "validate_schema",
    "assert_valid_schema",
    "assert_name",
    "assert_enum_value_name",
    "GraphQLType",
    "GraphQLInputType",
    "GraphQLOutputType",
    "GraphQLLeafType",
    "GraphQLCompositeType",
    "GraphQLAbstractType",
    "GraphQLWrappingType",
    "GraphQLNullableType",
    "GraphQLNamedType",
    "GraphQLNamedInputType",
    "GraphQLNamedOutputType",
    "Thunk",
    "ThunkCollection",
    "ThunkMapping",
    "GraphQLArgument",
    "GraphQLArgumentMap",
    "GraphQLEnumValue",
    "GraphQLEnumValueMap",
    "GraphQLField",
    "GraphQLFieldMap",
    "GraphQLFieldResolver",
    "GraphQLInputField",
    "GraphQLInputFieldMap",
    "GraphQLScalarSerializer",
    "GraphQLScalarValueParser",
    "GraphQLScalarLiteralParser",
    "GraphQLIsTypeOfFn",
    "GraphQLResolveInfo",
    "ResponsePath",
    "GraphQLTypeResolver",
    "GraphQLArgumentKwargs",
    "GraphQLDirectiveKwargs",
    "GraphQLEnumTypeKwargs",
    "GraphQLEnumValueKwargs",
    "GraphQLFieldKwargs",
    "GraphQLInputFieldKwargs",
    "GraphQLInputObjectTypeKwargs",
    "GraphQLInterfaceTypeKwargs",
    "GraphQLNamedTypeKwargs",
    "GraphQLObjectTypeKwargs",
    "GraphQLScalarTypeKwargs",
    "GraphQLSchemaKwargs",
    "GraphQLUnionTypeKwargs",
    "Source",
    "get_location",
    "print_location",
    "print_source_location",
    "Lexer",
    "TokenKind",
    "parse",
    "parse_value",
    "parse_const_value",
    "parse_type",
    "print_ast",
    "visit",
    "ParallelVisitor",
    "TypeInfoVisitor",
    "Visitor",
    "VisitorAction",
    "VisitorKeyMap",
    "BREAK",
    "SKIP",
    "REMOVE",
    "IDLE",
    "DirectiveLocation",
    "is_definition_node",
    "is_executable_definition_node",
    "is_selection_node",
    "is_value_node",
    "is_const_value_node",
    "is_type_node",
    "is_type_system_definition_node",
    "is_type_definition_node",
    "is_type_system_extension_node",
    "is_type_extension_node",
    "SourceLocation",
    "Location",
    "Token",
    "Node",
    "NameNode",
    "DocumentNode",
    "DefinitionNode",
    "ExecutableDefinitionNode",
    "OperationDefinitionNode",
    "OperationType",
    "VariableDefinitionNode",
    "VariableNode",
    "SelectionSetNode",
    "SelectionNode",
    "FieldNode",
    "ArgumentNode",
    "ConstArgumentNode",
    "FragmentSpreadNode",
    "InlineFragmentNode",
    "FragmentDefinitionNode",
    "ValueNode",
    "ConstValueNode",
    "IntValueNode",
    "FloatValueNode",
    "StringValueNode",
    "BooleanValueNode",
    "NullValueNode",
    "EnumValueNode",
    "ListValueNode",
    "ConstListValueNode",
    "ObjectValueNode",
    "ConstObjectValueNode",
    "ObjectFieldNode",
    "ConstObjectFieldNode",
    "DirectiveNode",
    "ConstDirectiveNode",
    "TypeNode",
    "NamedTypeNode",
    "ListTypeNode",
    "NonNullTypeNode",
    "TypeSystemDefinitionNode",
    "SchemaDefinitionNode",
    "OperationTypeDefinitionNode",
    "TypeDefinitionNode",
    "ScalarTypeDefinitionNode",
    "ObjectTypeDefinitionNode",
    "FieldDefinitionNode",
    "InputValueDefinitionNode",
    "InterfaceTypeDefinitionNode",
    "UnionTypeDefinitionNode",
    "EnumTypeDefinitionNode",
    "EnumValueDefinitionNode",
    "InputObjectTypeDefinitionNode",
    "DirectiveDefinitionNode",
    "TypeSystemExtensionNode",
    "SchemaExtensionNode",
    "TypeExtensionNode",
    "ScalarTypeExtensionNode",
    "ObjectTypeExtensionNode",
    "InterfaceTypeExtensionNode",
    "UnionTypeExtensionNode",
    "EnumTypeExtensionNode",
    "InputObjectTypeExtensionNode",
    "execute",
    "execute_sync",
    "default_field_resolver",
    "default_type_resolver",
    "get_argument_values",
    "get_directive_values",
    "get_variable_values",
    "ExecutionContext",
    "ExecutionResult",
    "FormattedExecutionResult",
    "Middleware",
    "MiddlewareManager",
    "subscribe",
    "create_source_event_stream",
    "MapAsyncIterator",
    "validate",
    "ValidationContext",
    "ValidationRule",
    "ASTValidationRule",
    "SDLValidationRule",
    "specified_rules",
    "ExecutableDefinitionsRule",
    "FieldsOnCorrectTypeRule",
    "FragmentsOnCompositeTypesRule",
    "KnownArgumentNamesRule",
    "KnownDirectivesRule",
    "KnownFragmentNamesRule",
    "KnownTypeNamesRule",
    "LoneAnonymousOperationRule",
    "NoFragmentCyclesRule",
    "NoUndefinedVariablesRule",
    "NoUnusedFragmentsRule",
    "NoUnusedVariablesRule",
    "OverlappingFieldsCanBeMergedRule",
    "PossibleFragmentSpreadsRule",
    "ProvidedRequiredArgumentsRule",
    "ScalarLeafsRule",
    "SingleFieldSubscriptionsRule",
    "UniqueArgumentNamesRule",
    "UniqueDirectivesPerLocationRule",
    "UniqueFragmentNamesRule",
    "UniqueInputFieldNamesRule",
    "UniqueOperationNamesRule",
    "UniqueVariableNamesRule",
    "ValuesOfCorrectTypeRule",
    "VariablesAreInputTypesRule",
    "VariablesInAllowedPositionRule",
    "LoneSchemaDefinitionRule",
    "UniqueOperationTypesRule",
    "UniqueTypeNamesRule",
    "UniqueEnumValueNamesRule",
    "UniqueFieldDefinitionNamesRule",
    "UniqueArgumentDefinitionNamesRule",
    "UniqueDirectiveNamesRule",
    "PossibleTypeExtensionsRule",
    "NoDeprecatedCustomRule",
    "NoSchemaIntrospectionCustomRule",
    "GraphQLError",
    "GraphQLErrorExtensions",
    "GraphQLFormattedError",
    "GraphQLSyntaxError",
    "located_error",
    "get_introspection_query",
    "IntrospectionQuery",
    "get_operation_ast",
    "get_operation_root_type",
    "introspection_from_schema",
    "build_client_schema",
    "build_ast_schema",
    "build_schema",
    "extend_schema",
    "lexicographic_sort_schema",
    "print_schema",
    "print_type",
    "print_introspection_schema",
    "type_from_ast",
    "value_from_ast",
    "value_from_ast_untyped",
    "ast_from_value",
    "ast_to_dict",
    "TypeInfo",
    "coerce_input_value",
    "concat_ast",
    "separate_operations",
    "strip_ignored_characters",
    "is_equal_type",
    "is_type_sub_type_of",
    "do_types_overlap",
    "assert_valid_name",
    "is_valid_name_error",
    "find_breaking_changes",
    "find_dangerous_changes",
    "BreakingChange",
    "BreakingChangeType",
    "DangerousChange",
    "DangerousChangeType",
    "Undefined",
    "UndefinedType",
]



from pytest import raises

from graphql.error import GraphQLError
from graphql.language import DirectiveLocation, DirectiveDefinitionNode, Node
from graphql.type import GraphQLArgument, GraphQLDirective, GraphQLInt, GraphQLString


def describe_type_system_directive():
    def can_create_instance():
        arg = GraphQLArgument(GraphQLString, description="arg description")
        node = DirectiveDefinitionNode()
        locations = [DirectiveLocation.SCHEMA, DirectiveLocation.OBJECT]
        directive = GraphQLDirective(
            name="test",
            locations=[DirectiveLocation.SCHEMA, DirectiveLocation.OBJECT],
            args={"arg": arg},
            description="test description",
            is_repeatable=True,
            ast_node=node,
        )
        assert directive.name == "test"
        assert directive.locations == tuple(locations)
        assert directive.args == {"arg": arg}
        assert directive.is_repeatable is True
        assert directive.description == "test description"
        assert directive.extensions == {}
        assert directive.ast_node is node

    def defines_a_directive_with_no_args():
        locations = [DirectiveLocation.QUERY]
        directive = GraphQLDirective("Foo", locations=locations)

        assert directive.name == "Foo"
        assert directive.args == {}
        assert directive.is_repeatable is False
        assert directive.extensions == {}
        assert directive.locations == tuple(locations)

    def defines_a_directive_with_multiple_args():
        args = {
            "foo": GraphQLArgument(GraphQLString),
            "bar": GraphQLArgument(GraphQLInt),
        }
        locations = [DirectiveLocation.QUERY]
        directive = GraphQLDirective("Foo", locations=locations, args=args)

        assert directive.name == "Foo"
        assert directive.args == args
        assert directive.is_repeatable is False
        assert directive.locations == tuple(locations)

    def defines_a_repeatable_directive():
        locations = [DirectiveLocation.QUERY]
        directive = GraphQLDirective("Foo", is_repeatable=True, locations=locations)

        assert directive.name == "Foo"
        assert directive.args == {}
        assert directive.is_repeatable is True
        assert directive.locations == tuple(locations)

    def directive_accepts_input_types_as_arguments():
        # noinspection PyTypeChecker
        directive = GraphQLDirective(
            name="Foo", locations=[], args={"arg": GraphQLString}  # type: ignore
        )
        arg = directive.args["arg"]
        assert isinstance(arg, GraphQLArgument)
        assert arg.type is GraphQLString

    def directive_accepts_strings_as_locations():
        # noinspection PyTypeChecker
        directive = GraphQLDirective(
            name="Foo", locations=["SCHEMA", "OBJECT"]  # type: ignore
        )
        assert directive.locations == (
            DirectiveLocation.SCHEMA,
            DirectiveLocation.OBJECT,
        )

    def directive_has_str():
        directive = GraphQLDirective("foo", [])
        assert str(directive) == "@foo"

    def directive_has_repr():
        directive = GraphQLDirective("foo", [])
        assert repr(directive) == "<GraphQLDirective(@foo)>"

    def can_compare_with_other_source_directive():
        locations = [DirectiveLocation.QUERY]
        directive = GraphQLDirective("Foo", locations)
        assert directive == directive
        assert not directive != directive
        assert not directive == {}
        assert directive != {}
        same_directive = GraphQLDirective("Foo", locations)
        assert directive == same_directive
        assert not directive != same_directive
        other_directive = GraphQLDirective("Bar", locations)
        assert not directive == other_directive
        assert directive != other_directive
        other_locations = [DirectiveLocation.MUTATION]
        other_directive = GraphQLDirective("Foo", other_locations)
        assert not directive == other_directive
        assert directive != other_directive
        other_directive = GraphQLDirective("Foo", locations, is_repeatable=True)
        assert not directive == other_directive
        assert directive != other_directive
        other_directive = GraphQLDirective("Foo", locations, description="other")
        assert not directive == other_directive
        assert directive != other_directive

    def rejects_a_directive_with_incorrectly_typed_name():
        with raises(TypeError, match="missing .* required .* 'name'"):
            # noinspection PyArgumentList
            GraphQLDirective()  # type: ignore
        with raises(TypeError) as exc_info:
            # noinspection PyTypeChecker
            GraphQLDirective(None, [])  # type: ignore
        assert str(exc_info.value) == "Must provide name."
        with raises(TypeError) as exc_info:
            # noinspection PyTypeChecker
            GraphQLDirective(42, {})  # type: ignore
        assert str(exc_info.value) == "Expected name to be a string."

    def rejects_a_directive_with_invalid_name():
        with raises(GraphQLError) as exc_info:
            GraphQLDirective("", [])
        assert str(exc_info.value) == "Expected name to be a non-empty string."
        with raises(GraphQLError) as exc_info:
            GraphQLDirective("bad-name", [])
        assert str(exc_info.value) == (
            "Names must only contain [_a-zA-Z0-9] but 'bad-name' does not."
        )

    def rejects_a_directive_with_incorrectly_typed_args():
        with raises(TypeError) as exc_info:
            # noinspection PyTypeChecker
            GraphQLDirective("Foo", locations=[], args=["arg"])  # type: ignore
        assert str(exc_info.value) == (
            "Foo args must be a dict with argument names as keys."
        )
        with raises(TypeError) as exc_info:
            # noinspection PyTypeChecker
            GraphQLDirective(
                "Foo",
                locations=[],
                args={1: GraphQLArgument(GraphQLString)},  # type: ignore
            )
        assert str(exc_info.value) == (
            "Foo args must be a dict with argument names as keys."
        )
        with raises(TypeError) as exc_info:
            # noinspection PyTypeChecker
            GraphQLDirective(
                "Foo",
                locations=[],
                args={"arg": GraphQLDirective("Bar", [])},  # type: ignore
            )
        assert str(exc_info.value) == (
            "Foo args must be GraphQLArgument or input type objects."
        )

    def rejects_a_directive_with_incorrectly_named_args():
        with raises(GraphQLError) as exc_info:
            GraphQLDirective(
                "Foo",
                locations=[DirectiveLocation.QUERY],
                args={"bad-name": GraphQLArgument(GraphQLString)},
            )
        assert str(exc_info.value) == (
            "Names must only contain [_a-zA-Z0-9] but 'bad-name' does not."
        )

    def rejects_a_directive_with_incorrectly_typed_repeatable_flag():
        with raises(TypeError) as exc_info:
            # noinspection PyTypeChecker
            GraphQLDirective("Foo", locations=[], is_repeatable=None)  # type: ignore
        assert str(exc_info.value) == "Foo is_repeatable flag must be True or False."

    def rejects_a_directive_with_undefined_locations():
        with raises(TypeError) as exc_info:
            # noinspection PyTypeChecker
            GraphQLDirective("Foo", locations=None)  # type: ignore
        assert str(exc_info.value) == (
            "Foo locations must be specified"
            " as a collection of DirectiveLocation enum values."
        )

    def rejects_a_directive_with_incorrectly_typed_locations():
        with raises(TypeError) as exc_info:
            # noinspection PyTypeChecker
            GraphQLDirective("Foo", locations="bad")  # type: ignore
        assert (
            str(exc_info.value) == "Foo locations must be specified"
            " as a collection of DirectiveLocation enum values."
        )
        with raises(TypeError) as exc_info:
            # noinspection PyTypeChecker
            GraphQLDirective("Foo", locations=["bad"])  # type: ignore
        assert str(exc_info.value) == (
            "Foo locations must be specified"
            " as a collection of DirectiveLocation enum values."
        )

    def rejects_a_directive_with_incorrectly_typed_description():
        with raises(TypeError) as exc_info:
            # noinspection PyTypeChecker
            GraphQLDirective(
                "Foo", locations=[], description={"bad": True}  # type: ignore
            )
        assert str(exc_info.value) == "Foo description must be a string."

    def rejects_a_directive_with_incorrectly_typed_ast_node():
        with raises(TypeError) as exc_info:
            # noinspection PyTypeChecker
            GraphQLDirective("Foo", locations=[], ast_node=Node())  # type: ignore
        assert str(exc_info.value) == (
            "Foo AST node must be a DirectiveDefinitionNode."
        )



from functools import partial

from graphql.language import parse
from graphql.type import GraphQLSchema
from graphql.utilities import build_schema
from graphql.validation import validate, FieldsOnCorrectTypeRule

from .harness import assert_validation_errors

test_schema = build_schema(
    """
    interface Pet {
      name: String
    }

    type Dog implements Pet {
      name: String
      nickname: String
      barkVolume: Int
    }

    type Cat implements Pet {
      name: String
      nickname: String
      meowVolume: Int
    }

    union CatOrDog = Cat | Dog

    type Human {
      name: String
      pets: [Pet]
    }

    type Query {
      human: Human
    }
    """
)


assert_errors = partial(
    assert_validation_errors, FieldsOnCorrectTypeRule, schema=test_schema
)

assert_valid = partial(assert_errors, errors=[])


def describe_validate_fields_on_correct_type():
    def object_field_selection():
        assert_valid(
            """
            fragment objectFieldSelection on Dog {
              __typename
              name
            }
            """
        )

    def aliased_object_field_selection():
        assert_valid(
            """
            fragment aliasedObjectFieldSelection on Dog {
              tn : __typename
              otherName : name
            }
            """
        )

    def interface_field_selection():
        assert_valid(
            """
            fragment interfaceFieldSelection on Pet {
              __typename
              name
            }
            """
        )

    def aliased_interface_field_selection():
        assert_valid(
            """
            fragment interfaceFieldSelection on Pet {
              otherName : name
            }
            """
        )

    def lying_alias_selection():
        assert_valid(
            """
            fragment lyingAliasSelection on Dog {
              name : nickname
            }
            """
        )

    def ignores_fields_on_unknown_type():
        assert_valid(
            """
            fragment unknownSelection on UnknownType {
              unknownField
            }
            """
        )

    def reports_errors_when_type_is_known_again():
        assert_errors(
            """
            fragment typeKnownAgain on Pet {
              unknown_pet_field {
                ... on Cat {
                  unknown_cat_field
                }
              }
            },
            """,
            [
                {
                    "message": "Cannot query field 'unknown_pet_field' on type 'Pet'.",
                    "locations": [(3, 15)],
                },
                {
                    "message": "Cannot query field 'unknown_cat_field' on type 'Cat'.",
                    "locations": [(5, 19)],
                },
            ],
        )

    def field_not_defined_on_fragment():
        assert_errors(
            """
            fragment fieldNotDefined on Dog {
              meowVolume
            }
            """,
            [
                {
                    "message": "Cannot query field 'meowVolume' on type 'Dog'."
                    " Did you mean 'barkVolume'?",
                    "locations": [(3, 15)],
                },
            ],
        )

    def ignores_deeply_unknown_field():
        assert_errors(
            """
            fragment deepFieldNotDefined on Dog {
              unknown_field {
                deeper_unknown_field
              }
            }
            """,
            [
                {
                    "message": "Cannot query field 'unknown_field' on type 'Dog'.",
                    "locations": [(3, 15)],
                },
            ],
        )

    def sub_field_not_defined():
        assert_errors(
            """
            fragment subFieldNotDefined on Human {
              pets {
                unknown_field
              }
            }
            """,
            [
                {
                    "message": "Cannot query field 'unknown_field' on type 'Pet'.",
                    "locations": [(4, 17)],
                },
            ],
        )

    def field_not_defined_on_inline_fragment():
        assert_errors(
            """
            fragment fieldNotDefined on Pet {
              ... on Dog {
                meowVolume
              }
            }
            """,
            [
                {
                    "message": "Cannot query field 'meowVolume' on type 'Dog'."
                    " Did you mean 'barkVolume'?",
                    "locations": [(4, 17)],
                },
            ],
        )

    def aliased_field_target_not_defined():
        assert_errors(
            """
            fragment aliasedFieldTargetNotDefined on Dog {
              volume : mooVolume
            }
            """,
            [
                {
                    "message": "Cannot query field 'mooVolume' on type 'Dog'."
                    " Did you mean 'barkVolume'?",
                    "locations": [(3, 15)],
                },
            ],
        )

    def aliased_lying_field_target_not_defined():
        assert_errors(
            """
            fragment aliasedLyingFieldTargetNotDefined on Dog {
              barkVolume : kawVolume
            }
            """,
            [
                {
                    "message": "Cannot query field 'kawVolume' on type 'Dog'."
                    " Did you mean 'barkVolume'?",
                    "locations": [(3, 15)],
                },
            ],
        )

    def not_defined_on_interface():
        assert_errors(
            """
            fragment notDefinedOnInterface on Pet {
              tailLength
            }
            """,
            [
                {
                    "message": "Cannot query field 'tailLength' on type 'Pet'.",
                    "locations": [(3, 15)],
                },
            ],
        )

    def defined_on_implementors_but_not_on_interface():
        assert_errors(
            """
            fragment definedOnImplementorsButNotInterface on Pet {
              nickname
            }
            """,
            [
                {
                    "message": "Cannot query field 'nickname' on type 'Pet'."
                    " Did you mean to use an inline fragment on 'Cat' or 'Dog'?",
                    "locations": [(3, 15)],
                },
            ],
        )

    def meta_field_selection_on_union():
        assert_valid(
            """
            fragment directFieldSelectionOnUnion on CatOrDog {
              __typename
            }
            """
        )

    def direct_field_selection_on_union():
        assert_errors(
            """
            fragment directFieldSelectionOnUnion on CatOrDog {
              directField
            }
            """,
            [
                {
                    "message": "Cannot query field 'directField' on type 'CatOrDog'.",
                    "locations": [(3, 15)],
                },
            ],
        )

    def defined_on_implementors_queried_on_union():
        assert_errors(
            """
              fragment definedOnImplementorsQueriedOnUnion on CatOrDog {
              name
            }
            """,
            [
                {
                    "message": "Cannot query field 'name' on type 'CatOrDog'."
                    " Did you mean to use an inline fragment"
                    " on 'Pet', 'Cat', or 'Dog'?",
                    "locations": [(3, 15)],
                },
            ],
        )

    def valid_field_in_inline_fragment():
        assert_valid(
            """
            fragment objectFieldSelection on Pet {
              ... on Dog {
                name
              }
              ... {
                name
              }
            }
            """
        )


def describe_fields_on_correct_type_error_message():
    def _error_message(schema: GraphQLSchema, query_str: str):
        errors = validate(schema, parse(query_str), [FieldsOnCorrectTypeRule])
        assert len(errors) == 1
        return errors[0].message

    def fields_correct_type_no_suggestion():
        schema = build_schema(
            """
            type T {
              fieldWithVeryLongNameThatWillNeverBeSuggested: String
            }
            type Query { t: T }
            """
        )
        assert _error_message(schema, "{ t { f } }") == (
            "Cannot query field 'f' on type 'T'."
        )

    def works_with_no_small_numbers_of_type_suggestion():
        schema = build_schema(
            """
            union T = A | B
            type Query { t: T }

            type A { f: String }
            type B { f: String }
            """
        )
        assert _error_message(schema, "{ t { f } }") == (
            "Cannot query field 'f' on type 'T'."
            " Did you mean to use an inline fragment on 'A' or 'B'?"
        )

    def works_with_no_small_numbers_of_field_suggestion():
        schema = build_schema(
            """
            type T {
              y: String
              z: String
            }
            type Query { t: T }
            """
        )
        assert _error_message(schema, "{ t { f } }") == (
            "Cannot query field 'f' on type 'T'. Did you mean 'y' or 'z'?"
        )

    def only_shows_one_set_of_suggestions_at_a_time_preferring_types():
        schema = build_schema(
            """
            interface T {
              y: String
              z: String
            }
            type Query { t: T }

            type A implements T {
              f: String
              y: String
              z: String
            }
            type B implements T {
              f: String
              y: String
              z: String
            }
            """
        )
        assert _error_message(schema, "{ t { f } }") == (
            "Cannot query field 'f' on type 'T'."
            " Did you mean to use an inline fragment on 'A' or 'B'?"
        )

    def sort_type_suggestions_based_on_inheritance_order():
        interface_schema = build_schema(
            """
            interface T { bar: String }
            type Query { t: T }

            interface Z implements T {
              foo: String
              bar: String
            }

            interface Y implements Z & T {
              foo: String
              bar: String
            }

            type X implements Y & Z & T {
              foo: String
              bar: String
            }
            """
        )

        assert _error_message(interface_schema, "{ t { foo } }") == (
            "Cannot query field 'foo' on type 'T'."
            " Did you mean to use an inline fragment on 'Z', 'Y', or 'X'?"
        )

        union_schema = build_schema(
            """
            interface Animal { name: String }
            interface Mammal implements Animal { name: String }

            interface Canine implements Animal & Mammal { name: String }
            type Dog implements Animal & Mammal & Canine { name: String }

            interface Feline implements Animal & Mammal { name: String }
            type Cat implements Animal & Mammal & Feline { name: String }

            union CatOrDog = Cat | Dog
            type Query { catOrDog: CatOrDog }
            """
        )

        assert _error_message(union_schema, "{ catOrDog { name } }") == (
            "Cannot query field 'name' on type 'CatOrDog'."
            " Did you mean to use an inline fragment"
            " on 'Animal', 'Mammal', 'Canine', 'Dog', or 'Feline'?"
        )

    def limits_lots_of_type_suggestions():
        schema = build_schema(
            """
            union T = A | B | C | D | E | F
            type Query { t: T }

            type A { f: String }
            type B { f: String }
            type C { f: String }
            type D { f: String }
            type E { f: String }
            type F { f: String }
            """
        )
        assert _error_message(schema, "{ t { f } }") == (
            "Cannot query field 'f' on type 'T'. Did you mean to use"
            " an inline fragment on 'A', 'B', 'C', 'D', or 'E'?"
        )

    def limits_lots_of_field_suggestions():
        schema = build_schema(
            """
            type T {
              u: String
              v: String
              w: String
              x: String
              y: String
              z: String
            }
            type Query { t: T }
            """
        )
        assert _error_message(schema, "{ t { f } }") == (
            "Cannot query field 'f' on type 'T'."
            " Did you mean 'u', 'v', 'w', 'x', or 'y'?"
        )



from functools import partial

from graphql.validation import UniqueOperationNamesRule

from .harness import assert_validation_errors

assert_errors = partial(assert_validation_errors, UniqueOperationNamesRule)

assert_valid = partial(assert_errors, errors=[])


def describe_validate_unique_operation_names():
    def no_operations():
        assert_valid(
            """
            fragment fragA on Type {
              field
            }
            """
        )

    def one_anon_operation():
        assert_valid(
            """
            {
              field
            }
            """
        )

    def one_named_operation():
        assert_valid(
            """
            query Foo {
              field
            }
            """
        )

    def multiple_operations():
        assert_valid(
            """
            query Foo {
              field
            }

            query Bar {
              field
            }
            """
        )

    def multiple_operations_of_different_types():
        assert_valid(
            """
            query Foo {
              field
            }

            mutation Bar {
              field
            }

            subscription Baz {
              field
            }
            """
        )

    def fragment_and_operation_named_the_same():
        assert_valid(
            """
            query Foo {
              ...Foo
            }
            fragment Foo on Type {
              field
            }
            """
        )

    def multiple_operations_of_same_name():
        assert_errors(
            """
            query Foo {
              fieldA
            }
            query Foo {
              fieldB
            }
            """,
            [
                {
                    "message": "There can be only one operation named 'Foo'.",
                    "locations": [(2, 19), (5, 19)],
                },
            ],
        )

    def multiple_ops_of_same_name_of_different_types_mutation():
        assert_errors(
            """
            query Foo {
              fieldA
            }
            mutation Foo {
              fieldB
            }
            """,
            [
                {
                    "message": "There can be only one operation named 'Foo'.",
                    "locations": [(2, 19), (5, 22)],
                },
            ],
        )

    def multiple_ops_of_same_name_of_different_types_subscription():
        assert_errors(
            """
            query Foo {
              fieldA
            }
            subscription Foo {
              fieldB
            }
            """,
            [
                {
                    "message": "There can be only one operation named 'Foo'.",
                    "locations": [(2, 19), (5, 26)],
                },
            ],
        )



from functools import partial

from graphql.utilities import build_schema
from graphql.validation.rules.possible_type_extensions import PossibleTypeExtensionsRule

from .harness import assert_sdl_validation_errors

assert_errors = partial(assert_sdl_validation_errors, PossibleTypeExtensionsRule)

assert_valid = partial(assert_errors, errors=[])


def describe_validate_possible_type_extensions():
    def no_extensions():
        assert_valid(
            """
            scalar FooScalar
            type FooObject
            interface FooInterface
            union FooUnion
            enum FooEnum
            input FooInputObject
            """
        )

    def one_extension_per_type():
        assert_valid(
            """
            scalar FooScalar
            type FooObject
            interface FooInterface
            union FooUnion
            enum FooEnum
            input FooInputObject

            extend scalar FooScalar @dummy
            extend type FooObject @dummy
            extend interface FooInterface @dummy
            extend union FooUnion @dummy
            extend enum FooEnum @dummy
            extend input FooInputObject @dummy
            """
        )

    def many_extensions_per_type():
        assert_valid(
            """
            scalar FooScalar
            type FooObject
            interface FooInterface
            union FooUnion
            enum FooEnum
            input FooInputObject

            extend scalar FooScalar @dummy
            extend type FooObject @dummy
            extend interface FooInterface @dummy
            extend union FooUnion @dummy
            extend enum FooEnum @dummy
            extend input FooInputObject @dummy

            extend scalar FooScalar @dummy
            extend type FooObject @dummy
            extend interface FooInterface @dummy
            extend union FooUnion @dummy
            extend enum FooEnum @dummy
            extend input FooInputObject @dummy
            """
        )

    def extending_unknown_type():
        message = (
            "Cannot extend type 'Unknown' because it is not defined."
            " Did you mean 'Known'?"
        )

        assert_errors(
            """
            type Known

            extend scalar Unknown @dummy
            extend type Unknown @dummy
            extend interface Unknown @dummy
            extend union Unknown @dummy
            extend enum Unknown @dummy
            extend input Unknown @dummy
            """,
            [
                {"message": message, "locations": [(4, 27)]},
                {"message": message, "locations": [(5, 25)]},
                {"message": message, "locations": [(6, 30)]},
                {"message": message, "locations": [(7, 26)]},
                {"message": message, "locations": [(8, 25)]},
                {"message": message, "locations": [(9, 26)]},
            ],
        )

    def does_not_consider_non_type_definitions():
        message = "Cannot extend type 'Foo' because it is not defined."

        assert_errors(
            """
            query Foo { __typename }
            fragment Foo on Query { __typename }
            directive @Foo on SCHEMA

            extend scalar Foo @dummy
            extend type Foo @dummy
            extend interface Foo @dummy
            extend union Foo @dummy
            extend enum Foo @dummy
            extend input Foo @dummy
            """,
            [
                {"message": message, "locations": [(6, 27)]},
                {"message": message, "locations": [(7, 25)]},
                {"message": message, "locations": [(8, 30)]},
                {"message": message, "locations": [(9, 26)]},
                {"message": message, "locations": [(10, 25)]},
                {"message": message, "locations": [(11, 26)]},
            ],
        )

    def extending_with_different_kinds():
        assert_errors(
            """
            scalar FooScalar
            type FooObject
            interface FooInterface
            union FooUnion
            enum FooEnum
            input FooInputObject

            extend type FooScalar @dummy
            extend interface FooObject @dummy
            extend union FooInterface @dummy
            extend enum FooUnion @dummy
            extend input FooEnum @dummy
            extend scalar FooInputObject @dummy
            """,
            [
                {
                    "message": "Cannot extend non-object type 'FooScalar'.",
                    "locations": [(2, 13), (9, 13)],
                },
                {
                    "message": "Cannot extend non-interface type 'FooObject'.",
                    "locations": [(3, 13), (10, 13)],
                },
                {
                    "message": "Cannot extend non-union type 'FooInterface'.",
                    "locations": [(4, 13), (11, 13)],
                },
                {
                    "message": "Cannot extend non-enum type 'FooUnion'.",
                    "locations": [(5, 13), (12, 13)],
                },
                {
                    "message": "Cannot extend non-input object type 'FooEnum'.",
                    "locations": [(6, 13), (13, 13)],
                },
                {
                    "message": "Cannot extend non-scalar type 'FooInputObject'.",
                    "locations": [(7, 13), (14, 13)],
                },
            ],
        )

    def extending_types_within_existing_schema():
        schema = build_schema(
            """
            scalar FooScalar
            type FooObject
            interface FooInterface
            union FooUnion
            enum FooEnum
            input FooInputObject
            """
        )
        sdl = """
            extend scalar FooScalar @dummy
            extend type FooObject @dummy
            extend interface FooInterface @dummy
            extend union FooUnion @dummy
            extend enum FooEnum @dummy
            extend input FooInputObject @dummy
            """

        assert_valid(sdl, schema=schema)

    def extending_unknown_types_within_existing_schema():
        schema = build_schema("type Known")
        sdl = """
            extend scalar Unknown @dummy
            extend type Unknown @dummy
            extend interface Unknown @dummy
            extend union Unknown @dummy
            extend enum Unknown @dummy
            extend input Unknown @dummy
            """

        message = (
            "Cannot extend type 'Unknown' because it is not defined."
            " Did you mean 'Known'?"
        )
        assert_errors(
            sdl,
            [
                {"message": message, "locations": [(2, 27)]},
                {"message": message, "locations": [(3, 25)]},
                {"message": message, "locations": [(4, 30)]},
                {"message": message, "locations": [(5, 26)]},
                {"message": message, "locations": [(6, 25)]},
                {"message": message, "locations": [(7, 26)]},
            ],
            schema,
        )

    def extending_types_with_different_kinds_within_existing_schema():
        schema = build_schema(
            """
            scalar FooScalar
            type FooObject
            interface FooInterface
            union FooUnion
            enum FooEnum
            input FooInputObject
            """
        )
        sdl = """
            extend type FooScalar @dummy
            extend interface FooObject @dummy
            extend union FooInterface @dummy
            extend enum FooUnion @dummy
            extend input FooEnum @dummy
            extend scalar FooInputObject @dummy
            """

        assert_errors(
            sdl,
            [
                {
                    "message": "Cannot extend non-object type 'FooScalar'.",
                    "locations": [(2, 13)],
                },
                {
                    "message": "Cannot extend non-interface type 'FooObject'.",
                    "locations": [(3, 13)],
                },
                {
                    "message": "Cannot extend non-union type 'FooInterface'.",
                    "locations": [(4, 13)],
                },
                {
                    "message": "Cannot extend non-enum type 'FooUnion'.",
                    "locations": [(5, 13)],
                },
                {
                    "message": "Cannot extend non-input object type 'FooEnum'.",
                    "locations": [(6, 13)],
                },
                {
                    "message": "Cannot extend non-scalar type 'FooInputObject'.",
                    "locations": [(7, 13)],
                },
            ],
            schema,
        )



from typing import Any, Dict, cast

from ...error import GraphQLError
from ...execution.collect_fields import collect_fields
from ...language import (
    FieldNode,
    FragmentDefinitionNode,
    OperationDefinitionNode,
    OperationType,
)
from . import ValidationRule

__all__ = ["SingleFieldSubscriptionsRule"]


class SingleFieldSubscriptionsRule(ValidationRule):
    """Subscriptions must only include a single non-introspection field.

    A GraphQL subscription is valid only if it contains a single root field and
    that root field is not an introspection field.

    See https://spec.graphql.org/draft/#sec-Single-root-field
    """

    def enter_operation_definition(
        self, node: OperationDefinitionNode, *_args: Any
    ) -> None:
        if node.operation != OperationType.SUBSCRIPTION:
            return
        schema = self.context.schema
        subscription_type = schema.subscription_type
        if subscription_type:
            operation_name = node.name.value if node.name else None
            variable_values: Dict[str, Any] = {}
            document = self.context.document
            fragments: Dict[str, FragmentDefinitionNode] = {
                definition.name.value: definition
                for definition in document.definitions
                if isinstance(definition, FragmentDefinitionNode)
            }
            fields = collect_fields(
                schema,
                fragments,
                variable_values,
                subscription_type,
                node.selection_set,
            )
            if len(fields) > 1:
                field_selection_lists = list(fields.values())
                extra_field_selection_lists = field_selection_lists[1:]
                extra_field_selection = [
                    field
                    for fields in extra_field_selection_lists
                    for field in (
                        fields
                        if isinstance(fields, list)
                        else [cast(FieldNode, fields)]
                    )
                ]
                self.report_error(
                    GraphQLError(
                        (
                            "Anonymous Subscription"
                            if operation_name is None
                            else f"Subscription '{operation_name}'"
                        )
                        + " must select only one top level field.",
                        extra_field_selection,
                    )
                )
            for field_nodes in fields.values():
                field = field_nodes[0]
                field_name = field.name.value
                if field_name.startswith("__"):
                    self.report_error(
                        GraphQLError(
                            (
                                "Anonymous Subscription"
                                if operation_name is None
                                else f"Subscription '{operation_name}'"
                            )
                            + " must not select an introspection top level field.",
                            field_nodes,
                        )
                    )



from functools import partial
from graphql.validation import UniqueArgumentNamesRule

from .harness import assert_validation_errors

assert_errors = partial(assert_validation_errors, UniqueArgumentNamesRule)

assert_valid = partial(assert_errors, errors=[])


def describe_validate_unique_argument_names():
    def no_arguments_on_field():
        assert_valid(
            """
            {
              field
            }
            """
        )

    def no_arguments_on_directive():
        assert_valid(
            """
            {
              field
            }
            """
        )

    def argument_on_field():
        assert_valid(
            """
            {
              field(arg: "value")
            }
            """
        )

    def argument_on_directive():
        assert_valid(
            """
            {
              field @directive(arg: "value")
            }
            """
        )

    def same_argument_on_two_fields():
        assert_valid(
            """
            {
              one: field(arg: "value")
              two: field(arg: "value")
            }
            """
        )

    def same_argument_on_field_and_directive():
        assert_valid(
            """
            {
              field(arg: "value") @directive(arg: "value")
            }
            """
        )

    def same_argument_on_two_directives():
        assert_valid(
            """
            {
              field @directive1(arg: "value") @directive2(arg: "value")
            }
            """
        )

    def multiple_field_arguments():
        assert_valid(
            """
            {
              field(arg1: "value", arg2: "value", arg3: "value")
            }
            """
        )

    def multiple_directive_arguments():
        assert_valid(
            """
            {
              field @directive(arg1: "value", arg2: "value", arg3: "value")
            }
            """
        )

    def duplicate_field_arguments():
        assert_errors(
            """
            {
              field(arg1: "value", arg1: "value")
            }
            """,
            [
                {
                    "message": "There can be only one argument named 'arg1'.",
                    "locations": [(3, 21), (3, 36)],
                },
            ],
        )

    def many_duplicate_field_arguments():
        assert_errors(
            """
            {
              field(arg1: "value", arg1: "value", arg1: "value")
            }
            """,
            [
                {
                    "message": "There can be only one argument named 'arg1'.",
                    "locations": [(3, 21), (3, 36), (3, 51)],
                },
            ],
        )

    def duplicate_directive_arguments():
        assert_errors(
            """
            {
              field @directive(arg1: "value", arg1: "value")
            }
            """,
            [
                {
                    "message": "There can be only one argument named 'arg1'.",
                    "locations": [(3, 32), (3, 47)],
                },
            ],
        )

    def many_duplicate_directive_arguments():
        assert_errors(
            """
            {
              field @directive(arg1: "value", arg1: "value", arg1: "value")
            }
            """,
            [
                {
                    "message": "There can be only one argument named 'arg1'.",
                    "locations": [(3, 32), (3, 47), (3, 62)],
                },
            ],
        )



from graphql import build_client_schema, GraphQLSchema

from ..fixtures import big_schema_introspection_result  # noqa: F401


def test_build_schema_from_introspection(
    benchmark, big_schema_introspection_result  # noqa: F811
):
    schema: GraphQLSchema = benchmark(
        lambda: build_client_schema(
            big_schema_introspection_result["data"], assume_valid=True
        )
    )
    assert schema.query_type is not None



from math import isnan, nan
from typing import Any, Dict, Optional

from graphql.language import parse_value, ValueNode
from graphql.pyutils import Undefined
from graphql.type import (
    GraphQLBoolean,
    GraphQLEnumType,
    GraphQLFloat,
    GraphQLID,
    GraphQLInputField,
    GraphQLInputObjectType,
    GraphQLInputType,
    GraphQLInt,
    GraphQLList,
    GraphQLNonNull,
    GraphQLScalarType,
    GraphQLString,
)
from graphql.utilities import value_from_ast


def describe_value_from_ast():
    def _value_from(
        value_text: str,
        type_: GraphQLInputType,
        variables: Optional[Dict[str, Any]] = None,
    ):
        ast = parse_value(value_text)
        return value_from_ast(ast, type_, variables)

    def rejects_empty_input():
        # noinspection PyTypeChecker
        assert value_from_ast(None, GraphQLBoolean) is Undefined

    def converts_according_to_input_coercion_rules():
        assert _value_from("true", GraphQLBoolean) is True
        assert _value_from("false", GraphQLBoolean) is False
        assert _value_from("123", GraphQLInt) == 123
        assert _value_from("123", GraphQLFloat) == 123
        assert _value_from("123.456", GraphQLFloat) == 123.456
        assert _value_from('"abc123"', GraphQLString) == "abc123"
        assert _value_from("123456", GraphQLID) == "123456"
        assert _value_from('"123456"', GraphQLID) == "123456"

    def does_not_convert_when_input_coercion_rules_reject_a_value():
        assert _value_from("123", GraphQLBoolean) is Undefined
        assert _value_from("123.456", GraphQLInt) is Undefined
        assert _value_from("true", GraphQLInt) is Undefined
        assert _value_from('"123"', GraphQLInt) is Undefined
        assert _value_from('"123"', GraphQLFloat) is Undefined
        assert _value_from("123", GraphQLString) is Undefined
        assert _value_from("true", GraphQLString) is Undefined
        assert _value_from("123.456", GraphQLID) is Undefined

    def convert_using_parse_literal_from_a_custom_scalar_type():
        def pass_through_parse_literal(node, _vars=None):
            assert node.kind == "string_value"
            return node.value

        pass_through_scalar = GraphQLScalarType(
            "PassThroughScalar",
            parse_literal=pass_through_parse_literal,
            parse_value=lambda value: value,  # pragma: no cover
        )

        assert _value_from('"value"', pass_through_scalar) == "value"

        def throw_parse_literal(_node: ValueNode, _vars=None):
            raise RuntimeError("Test")

        throw_scalar = GraphQLScalarType(
            "ThrowScalar",
            parse_literal=throw_parse_literal,
            parse_value=lambda value: value,  # pragma: no cover
        )

        assert _value_from("value", throw_scalar) is Undefined

        def undefined_parse_literal(_node: ValueNode, _vars=None):
            return Undefined

        return_undefined_scalar = GraphQLScalarType(
            "ReturnUndefinedScalar",
            parse_literal=undefined_parse_literal,
            parse_value=lambda value: value,  # pragma: no cover
        )

        assert _value_from("value", return_undefined_scalar) is Undefined

    def converts_enum_values_according_to_input_coercion_rules():
        test_enum = GraphQLEnumType(
            "TestColor",
            {
                "RED": 1,
                "GREEN": 2,
                "BLUE": 3,
                "NULL": None,
                "NAN": nan,
                "NO_CUSTOM_VALUE": Undefined,
            },
        )

        assert _value_from("RED", test_enum) == 1
        assert _value_from("BLUE", test_enum) == 3
        assert _value_from("YELLOW", test_enum) is Undefined
        assert _value_from("3", test_enum) is Undefined
        assert _value_from('"BLUE"', test_enum) is Undefined
        assert _value_from("null", test_enum) is None
        assert _value_from("NULL", test_enum) is None
        assert _value_from("NULL", GraphQLNonNull(test_enum)) is None
        assert isnan(_value_from("NAN", test_enum))
        assert _value_from("NO_CUSTOM_VALUE", test_enum) is Undefined

    # Boolean!
    non_null_bool = GraphQLNonNull(GraphQLBoolean)
    # [Boolean]
    list_of_bool = GraphQLList(GraphQLBoolean)
    # [Boolean!]
    list_of_non_null_bool = GraphQLList(non_null_bool)
    # [Boolean]!
    non_null_list_of_bool = GraphQLNonNull(list_of_bool)
    # [Boolean!]!
    non_null_list_of_non_mull_bool = GraphQLNonNull(list_of_non_null_bool)

    def coerces_to_null_unless_non_null():
        assert _value_from("null", GraphQLBoolean) is None
        assert _value_from("null", non_null_bool) is Undefined

    def coerces_lists_of_values():
        assert _value_from("true", list_of_bool) == [True]
        assert _value_from("123", list_of_bool) is Undefined
        assert _value_from("null", list_of_bool) is None
        assert _value_from("[true, false]", list_of_bool) == [True, False]
        assert _value_from("[true, 123]", list_of_bool) is Undefined
        assert _value_from("[true, null]", list_of_bool) == [True, None]
        assert _value_from("{ true: true }", list_of_bool) is Undefined

    def coerces_non_null_lists_of_values():
        assert _value_from("true", non_null_list_of_bool) == [True]
        assert _value_from("123", non_null_list_of_bool) is Undefined
        assert _value_from("null", non_null_list_of_bool) is Undefined
        assert _value_from("[true, false]", non_null_list_of_bool) == [True, False]
        assert _value_from("[true, 123]", non_null_list_of_bool) is Undefined
        assert _value_from("[true, null]", non_null_list_of_bool) == [True, None]

    def coerces_lists_of_non_null_values():
        assert _value_from("true", list_of_non_null_bool) == [True]
        assert _value_from("123", list_of_non_null_bool) is Undefined
        assert _value_from("null", list_of_non_null_bool) is None
        assert _value_from("[true, false]", list_of_non_null_bool) == [True, False]
        assert _value_from("[true, 123]", list_of_non_null_bool) is Undefined
        assert _value_from("[true, null]", list_of_non_null_bool) is Undefined

    def coerces_non_null_lists_of_non_null_values():
        assert _value_from("true", non_null_list_of_non_mull_bool) == [True]
        assert _value_from("123", non_null_list_of_non_mull_bool) is Undefined
        assert _value_from("null", non_null_list_of_non_mull_bool) is Undefined
        assert _value_from("[true, false]", non_null_list_of_non_mull_bool) == [
            True,
            False,
        ]
        assert _value_from("[true, 123]", non_null_list_of_non_mull_bool) is Undefined
        assert _value_from("[true, null]", non_null_list_of_non_mull_bool) is Undefined

    test_input_obj = GraphQLInputObjectType(
        "TestInput",
        {
            "int": GraphQLInputField(GraphQLInt, default_value=42),
            "bool": GraphQLInputField(GraphQLBoolean),
            "requiredBool": GraphQLInputField(non_null_bool),
        },
    )

    def coerces_input_objects_according_to_input_coercion_rules():
        assert _value_from("null", test_input_obj) is None
        assert _value_from("[]", test_input_obj) is Undefined
        assert _value_from("123", test_input_obj) is Undefined
        assert _value_from("{ int: 123, requiredBool: false }", test_input_obj) == {
            "int": 123,
            "requiredBool": False,
        }
        assert _value_from("{ bool: true, requiredBool: false }", test_input_obj) == {
            "int": 42,
            "bool": True,
            "requiredBool": False,
        }
        assert (
            _value_from("{ int: true, requiredBool: true }", test_input_obj)
            is Undefined
        )
        assert _value_from("{ requiredBool: null }", test_input_obj) is Undefined
        assert _value_from("{ bool: true }", test_input_obj) is Undefined

    def accepts_variable_values_assuming_already_coerced():
        assert _value_from("$var", GraphQLBoolean, {}) is Undefined
        assert _value_from("$var", GraphQLBoolean, {"var": True}) is True
        assert _value_from("$var", GraphQLBoolean, {"var": None}) is None
        assert _value_from("$var", non_null_bool, {"var": None}) is Undefined

    def asserts_variables_are_provided_as_items_in_lists():
        assert _value_from("[ $foo ]", list_of_bool, {}) == [None]
        assert _value_from("[ $foo ]", list_of_non_null_bool, {}) is Undefined
        assert _value_from("[ $foo ]", list_of_non_null_bool, {"foo": True}) == [True]
        # Note: variables are expected to have already been coerced, so we
        # do not expect the singleton wrapping behavior for variables.
        assert _value_from("$foo", list_of_non_null_bool, {"foo": True}) is True
        assert _value_from("$foo", list_of_non_null_bool, {"foo": [True]}) == [True]

    def omits_input_object_fields_for_unprovided_variables():
        assert _value_from(
            "{ int: $foo, bool: $foo, requiredBool: true }", test_input_obj, {}
        ) == {"int": 42, "requiredBool": True}
        assert _value_from("{ requiredBool: $foo }", test_input_obj, {}) is Undefined
        assert _value_from("{ requiredBool: $foo }", test_input_obj, {"foo": True}) == {
            "int": 42,
            "requiredBool": True,
        }

    def transforms_names_using_out_name():
        # This is an extension of GraphQL.js.
        complex_input_obj = GraphQLInputObjectType(
            "Complex",
            {
                "realPart": GraphQLInputField(GraphQLFloat, out_name="real_part"),
                "imagPart": GraphQLInputField(
                    GraphQLFloat, default_value=0, out_name="imag_part"
                ),
            },
        )
        assert _value_from("{ realPart: 1 }", complex_input_obj) == {
            "real_part": 1,
            "imag_part": 0,
        }

    def transforms_values_with_out_type():
        # This is an extension of GraphQL.js.
        complex_input_obj = GraphQLInputObjectType(
            "Complex",
            {
                "real": GraphQLInputField(GraphQLFloat),
                "imag": GraphQLInputField(GraphQLFloat),
            },
            out_type=lambda value: complex(value["real"], value["imag"]),
        )
        assert _value_from("{ real: 1, imag: 2 }", complex_input_obj) == 1 + 2j



from typing import List, Optional, Tuple

from pytest import raises

from graphql.error import GraphQLSyntaxError
from graphql.language import Lexer, Source, SourceLocation, Token, TokenKind
from graphql.language.lexer import is_punctuator_token_kind
from graphql.pyutils import inspect

from ..utils import dedent

Location = Optional[Tuple[int, int]]


def lex_one(s: str) -> Token:
    lexer = Lexer(Source(s))
    return lexer.advance()


def lex_second(s: str) -> Token:
    lexer = Lexer(Source(s))
    lexer.advance()
    return lexer.advance()


def assert_syntax_error(text: str, message: str, location: Location) -> None:
    with raises(GraphQLSyntaxError) as exc_info:
        lex_second(text)
    error = exc_info.value
    assert error.message == f"Syntax Error: {message}"
    assert error.description == message
    assert error.locations == [location]


def describe_lexer():
    def ignores_bom_header():
        token = lex_one("\uFEFF foo")
        assert token == Token(TokenKind.NAME, 2, 5, 1, 3, "foo")

    def tracks_line_breaks():
        assert lex_one("foo") == Token(TokenKind.NAME, 0, 3, 1, 1, "foo")
        assert lex_one("\nfoo") == Token(TokenKind.NAME, 1, 4, 2, 1, "foo")
        assert lex_one("\rfoo") == Token(TokenKind.NAME, 1, 4, 2, 1, "foo")
        assert lex_one("\r\nfoo") == Token(TokenKind.NAME, 2, 5, 2, 1, "foo")
        assert lex_one("\n\rfoo") == Token(TokenKind.NAME, 2, 5, 3, 1, "foo")
        assert lex_one("\r\r\n\nfoo") == Token(TokenKind.NAME, 4, 7, 4, 1, "foo")
        assert lex_one("\n\n\r\rfoo") == Token(TokenKind.NAME, 4, 7, 5, 1, "foo")

    def records_line_and_column():
        token = lex_one("\n \r\n \r  foo\n")
        assert token == Token(TokenKind.NAME, 8, 11, 4, 3, "foo")

    def can_be_stringified_or_pyutils_inspected():
        token = lex_one("foo")
        assert token.desc == "Name 'foo'"
        assert str(token) == token.desc
        assert repr(token) == "<Token Name 'foo' 1:1>"
        assert inspect(token) == repr(token)

    def skips_whitespace_and_comments():
        token = lex_one("\n\n    foo\n\n\n")
        assert token == Token(TokenKind.NAME, 6, 9, 3, 5, "foo")
        token = lex_one("\r\n\r\n  foo\r\n\r\n")
        assert token == Token(TokenKind.NAME, 6, 9, 3, 3, "foo")
        token = lex_one("\r\r  foo\r\r")
        assert token == Token(TokenKind.NAME, 4, 7, 3, 3, "foo")
        token = lex_one("\t\tfoo\t\t")
        assert token == Token(TokenKind.NAME, 2, 5, 1, 3, "foo")
        token = lex_one("\n    #comment\n    foo#comment\n")
        assert token == Token(TokenKind.NAME, 18, 21, 3, 5, "foo")
        token = lex_one(",,,foo,,,")
        assert token == Token(TokenKind.NAME, 3, 6, 1, 4, "foo")

    def errors_respect_whitespace():
        with raises(GraphQLSyntaxError) as exc_info:
            lex_one("\n\n ~\n")

        assert str(exc_info.value) == dedent(
            """
            Syntax Error: Unexpected character: '~'.

            GraphQL request:3:2
            2 |
            3 |  ~
              |  ^
            4 |
            """
        )

    def updates_line_numbers_in_error_for_file_context():
        s = "\n\n     ~\n\n"
        source = Source(s, "foo.js", SourceLocation(11, 12))
        with raises(GraphQLSyntaxError) as exc_info:
            Lexer(source).advance()
        assert str(exc_info.value) == dedent(
            """
            Syntax Error: Unexpected character: '~'.

            foo.js:13:6
            12 |
            13 |      ~
               |      ^
            14 |
            """
        )

    def updates_column_numbers_in_error_for_file_context():
        source = Source("~", "foo.js", SourceLocation(1, 5))
        with raises(GraphQLSyntaxError) as exc_info:
            Lexer(source).advance()
        assert str(exc_info.value) == dedent(
            """
            Syntax Error: Unexpected character: '~'.

            foo.js:1:5
            1 |     ~
              |     ^
            """
        )

    # noinspection PyArgumentEqualDefault
    def lexes_empty_string():
        token = lex_one('""')
        assert token == Token(TokenKind.STRING, 0, 2, 1, 1, "")
        assert token.value == ""

    # noinspection PyArgumentEqualDefault
    def lexes_strings():
        assert lex_one('""') == Token(TokenKind.STRING, 0, 2, 1, 1, "")
        assert lex_one('"simple"') == Token(TokenKind.STRING, 0, 8, 1, 1, "simple")
        assert lex_one('" white space "') == Token(
            TokenKind.STRING, 0, 15, 1, 1, " white space "
        )
        assert lex_one('"quote \\""') == Token(TokenKind.STRING, 0, 10, 1, 1, 'quote "')
        assert lex_one('"escaped \\n\\r\\b\\t\\f"') == Token(
            TokenKind.STRING, 0, 20, 1, 1, "escaped \n\r\b\t\f"
        )
        assert lex_one('"slashes \\\\ \\/"') == Token(
            TokenKind.STRING, 0, 15, 1, 1, "slashes \\ /"
        )
        assert lex_one('"unescaped surrogate pair \uD83D\uDE00"') == Token(
            TokenKind.STRING, 0, 29, 1, 1, "unescaped surrogate pair \uD83D\uDE00"
        )
        assert lex_one('"unescaped unicode outside BMP \U0001f600"') == Token(
            TokenKind.STRING, 0, 33, 1, 1, "unescaped unicode outside BMP \U0001f600"
        )
        assert lex_one('"unescaped maximal unicode outside BMP \U0010ffff"') == Token(
            TokenKind.STRING,
            0,
            41,
            1,
            1,
            "unescaped maximal unicode outside BMP \U0010ffff",
        )
        assert lex_one('"unicode \\u1234\\u5678\\u90AB\\uCDEF"') == Token(
            TokenKind.STRING, 0, 34, 1, 1, "unicode \u1234\u5678\u90AB\uCDEF"
        )
        assert lex_one('"unicode \\u{1234}\\u{5678}\\u{90AB}\\u{CDEF}"') == Token(
            TokenKind.STRING, 0, 42, 1, 1, "unicode \u1234\u5678\u90AB\uCDEF"
        )
        assert lex_one('"string with unicode escape outside BMP \\u{1F600}"') == Token(
            TokenKind.STRING,
            0,
            50,
            1,
            1,
            "string with unicode escape outside BMP \U0001F600",
        )
        assert lex_one('"string with minimal unicode escape \\u{0}"') == Token(
            TokenKind.STRING, 0, 42, 1, 1, "string with minimal unicode escape \u0000"
        )
        assert lex_one('"string with maximal unicode escape \\u{10FFFF}"') == Token(
            TokenKind.STRING,
            0,
            47,
            1,
            1,
            "string with maximal unicode escape \U0010FFFF",
        )
        assert lex_one(
            '"string with maximal minimal unicode escape \\u{00000000}"'
        ) == Token(
            TokenKind.STRING,
            0,
            57,
            1,
            1,
            "string with maximal minimal unicode escape \u0000",
        )
        assert lex_one(
            '"string with unicode surrogate pair escape \\uD83D\\uDE00"'
        ) == Token(
            TokenKind.STRING,
            0,
            56,
            1,
            1,
            "string with unicode surrogate pair escape \U0001f600",
        )
        assert lex_one(
            '"string with unicode surrogate pair escape \\uD800\\uDC00"'
        ) == Token(
            TokenKind.STRING,
            0,
            56,
            1,
            1,
            "string with unicode surrogate pair escape \U00010000",
        )
        assert lex_one(
            '"string with unicode surrogate pair escape \\uDBFF\\uDFFF"'
        ) == Token(
            TokenKind.STRING,
            0,
            56,
            1,
            1,
            "string with unicode surrogate pair escape \U0010FFFF",
        )

    def lex_reports_useful_string_errors():
        assert_syntax_error('"', "Unterminated string.", (1, 2))
        assert_syntax_error('"""', "Unterminated string.", (1, 4))
        assert_syntax_error('""""', "Unterminated string.", (1, 5))
        assert_syntax_error('"no end quote', "Unterminated string.", (1, 14))
        assert_syntax_error(
            "'single quotes'",
            "Unexpected single quote character ('), "
            'did you mean to use a double quote (")?',
            (1, 1),
        )
        assert_syntax_error(
            '"bad surrogate \uDEAD"',
            "Invalid character within String: U+DEAD.",
            (1, 16),
        )
        assert_syntax_error(
            '"bad high surrogate pair \uDEAD\uDEAD"',
            "Invalid character within String: U+DEAD.",
            (1, 26),
        )
        assert_syntax_error(
            '"bad low surrogate pair \uD800\uD800"',
            "Invalid character within String: U+D800.",
            (1, 25),
        )
        assert_syntax_error('"multi\nline"', "Unterminated string.", (1, 7))
        assert_syntax_error('"multi\rline"', "Unterminated string.", (1, 7))
        assert_syntax_error(
            '"bad \\z esc"', "Invalid character escape sequence: '\\z'.", (1, 6)
        )
        assert_syntax_error(
            '"bad \\x esc"', "Invalid character escape sequence: '\\x'.", (1, 6)
        )
        assert_syntax_error(
            '"bad \\u1 esc"', "Invalid Unicode escape sequence: '\\u1 es'.", (1, 6)
        )
        assert_syntax_error(
            '"bad \\u0XX1 esc"', "Invalid Unicode escape sequence: '\\u0XX1'.", (1, 6)
        )
        assert_syntax_error(
            '"bad \\uXXXX esc"', "Invalid Unicode escape sequence: '\\uXXXX'.", (1, 6)
        )
        assert_syntax_error(
            '"bad \\uFXXX esc"', "Invalid Unicode escape sequence: '\\uFXXX'.", (1, 6)
        )
        assert_syntax_error(
            '"bad \\uXXXF esc"', "Invalid Unicode escape sequence: '\\uXXXF'.", (1, 6)
        )
        assert_syntax_error(
            '"bad \\u{} esc"', "Invalid Unicode escape sequence: '\\u{}'.", (1, 6)
        )
        assert_syntax_error(
            '"bad \\u{FXXX} esc"', "Invalid Unicode escape sequence: '\\u{FX'.", (1, 6)
        )
        assert_syntax_error(
            '"bad \\u{FFFF esc"',
            "Invalid Unicode escape sequence: '\\u{FFFF '.",
            (1, 6),
        )
        assert_syntax_error(
            '"bad \\u{FFFF"', "Invalid Unicode escape sequence: '\\u{FFFF\"'.", (1, 6)
        )
        assert_syntax_error(
            '"too high \\u{110000} esc"',
            "Invalid Unicode escape sequence: '\\u{110000}'.",
            (1, 11),
        )
        assert_syntax_error(
            '"way too high \\u{12345678} esc"',
            "Invalid Unicode escape sequence: '\\u{12345678}'.",
            (1, 15),
        )
        assert_syntax_error(
            '"too long \\u{000000000} esc"',
            "Invalid Unicode escape sequence: '\\u{000000000'.",
            (1, 11),
        )
        assert_syntax_error(
            '"bad surrogate \\uDEAD esc"',
            "Invalid Unicode escape sequence: '\\uDEAD'.",
            (1, 16),
        )
        assert_syntax_error(
            '"bad surrogate \\u{DEAD} esc"',
            "Invalid Unicode escape sequence: '\\u{DEAD}'.",
            (1, 16),
        )
        assert_syntax_error(
            '"cannot use braces for surrogate pair \\u{D83D}\\u{DE00} esc"',
            "Invalid Unicode escape sequence: '\\u{D83D}'.",
            (1, 39),
        )
        assert_syntax_error(
            '"bad high surrogate pair \\uDEAD\\uDEAD esc"',
            "Invalid Unicode escape sequence: '\\uDEAD'.",
            (1, 26),
        )
        assert_syntax_error(
            '"bad low surrogate pair \\uD800\\uD800 esc"',
            "Invalid Unicode escape sequence: '\\uD800'.",
            (1, 25),
        )
        assert_syntax_error(
            '"cannot escape half a pair \uD83D\\uDE00 esc"',
            "Invalid character within String: U+D83D.",
            (1, 28),
        )
        assert_syntax_error(
            '"cannot escape half a pair \\uD83D\uDE00 esc"',
            "Invalid Unicode escape sequence: '\\uD83D'.",
            (1, 28),
        )
        assert_syntax_error(
            '"bad \\uD83D\\not an escape"',
            "Invalid Unicode escape sequence: '\\uD83D'.",
            (1, 6),
        )

    # noinspection PyArgumentEqualDefault
    def lexes_block_strings():
        assert lex_one('""""""') == Token(TokenKind.BLOCK_STRING, 0, 6, 1, 1, "")
        assert lex_one('"""simple"""') == Token(
            TokenKind.BLOCK_STRING, 0, 12, 1, 1, "simple"
        )
        assert lex_one('""" white space """') == Token(
            TokenKind.BLOCK_STRING, 0, 19, 1, 1, " white space "
        )
        assert lex_one('"""contains " quote"""') == Token(
            TokenKind.BLOCK_STRING, 0, 22, 1, 1, 'contains " quote'
        )
        assert lex_one('"""contains \\""" triple-quote"""') == Token(
            TokenKind.BLOCK_STRING, 0, 32, 1, 1, 'contains """ triple-quote'
        )
        assert lex_one('"""multi\nline"""') == Token(
            TokenKind.BLOCK_STRING, 0, 16, 1, 1, "multi\nline"
        )
        assert lex_one('"""multi\rline\r\nnormalized"""') == Token(
            TokenKind.BLOCK_STRING, 0, 28, 1, 1, "multi\nline\nnormalized"
        )
        assert lex_one('"""unescaped \\n\\r\\b\\t\\f\\u1234"""') == Token(
            TokenKind.BLOCK_STRING,
            0,
            32,
            1,
            1,
            "unescaped \\n\\r\\b\\t\\f\\u1234",
        )
        assert lex_one('"""unescaped surrogate pair \uD83D\uDE00"""') == Token(
            TokenKind.BLOCK_STRING,
            0,
            33,
            1,
            1,
            "unescaped surrogate pair \uD83D\uDE00",
        )
        assert lex_one('"""unescaped unicode outside BMP \U0001f600"""') == Token(
            TokenKind.BLOCK_STRING,
            0,
            37,
            1,
            1,
            "unescaped unicode outside BMP \U0001f600",
        )
        assert lex_one('"""slashes \\\\ \\/"""') == Token(
            TokenKind.BLOCK_STRING, 0, 19, 1, 1, "slashes \\\\ \\/"
        )
        assert lex_one(
            '"""\n\n        spans\n          multiple\n'
            '            lines\n\n        """'
        ) == Token(TokenKind.BLOCK_STRING, 0, 68, 1, 1, "spans\n  multiple\n    lines")

    def advance_line_after_lexing_multiline_block_string():
        assert (
            lex_second(
                '''"""

        spans
          multiple
            lines

        \n """ second_token'''
            )
            == Token(TokenKind.NAME, 71, 83, 8, 6, "second_token")
        )

    def lex_reports_useful_block_string_errors():
        assert_syntax_error('"""', "Unterminated string.", (1, 4))
        assert_syntax_error('"""no end quote', "Unterminated string.", (1, 16))
        assert_syntax_error(
            '"""contains invalid surrogate \uDEAD"""',
            "Invalid character within String: U+DEAD.",
            (1, 31),
        )

    # noinspection PyArgumentEqualDefault
    def lexes_numbers():
        assert lex_one("0") == Token(TokenKind.INT, 0, 1, 1, 1, "0")
        assert lex_one("1") == Token(TokenKind.INT, 0, 1, 1, 1, "1")
        assert lex_one("4") == Token(TokenKind.INT, 0, 1, 1, 1, "4")
        assert lex_one("9") == Token(TokenKind.INT, 0, 1, 1, 1, "9")
        assert lex_one("42") == Token(TokenKind.INT, 0, 2, 1, 1, "42")
        assert lex_one("4.123") == Token(TokenKind.FLOAT, 0, 5, 1, 1, "4.123")
        assert lex_one("-4") == Token(TokenKind.INT, 0, 2, 1, 1, "-4")
        assert lex_one("-42") == Token(TokenKind.INT, 0, 3, 1, 1, "-42")
        assert lex_one("-4.123") == Token(TokenKind.FLOAT, 0, 6, 1, 1, "-4.123")
        assert lex_one("0.123") == Token(TokenKind.FLOAT, 0, 5, 1, 1, "0.123")
        assert lex_one("123e4") == Token(TokenKind.FLOAT, 0, 5, 1, 1, "123e4")
        assert lex_one("123E4") == Token(TokenKind.FLOAT, 0, 5, 1, 1, "123E4")
        assert lex_one("123e-4") == Token(TokenKind.FLOAT, 0, 6, 1, 1, "123e-4")
        assert lex_one("123e+4") == Token(TokenKind.FLOAT, 0, 6, 1, 1, "123e+4")
        assert lex_one("-1.123e4") == Token(TokenKind.FLOAT, 0, 8, 1, 1, "-1.123e4")
        assert lex_one("-1.123E4") == Token(TokenKind.FLOAT, 0, 8, 1, 1, "-1.123E4")
        assert lex_one("-1.123e-4") == Token(TokenKind.FLOAT, 0, 9, 1, 1, "-1.123e-4")
        assert lex_one("-1.123e+4") == Token(TokenKind.FLOAT, 0, 9, 1, 1, "-1.123e+4")
        assert lex_one("-1.123e4567") == Token(
            TokenKind.FLOAT, 0, 11, 1, 1, "-1.123e4567"
        )

    def lex_reports_useful_number_errors():
        assert_syntax_error(
            "00", "Invalid number, unexpected digit after 0: '0'.", (1, 2)
        )
        assert_syntax_error(
            "01", "Invalid number, unexpected digit after 0: '1'.", (1, 2)
        )
        assert_syntax_error(
            "01.23", "Invalid number, unexpected digit after 0: '1'.", (1, 2)
        )
        assert_syntax_error("+1", "Unexpected character: '+'.", (1, 1))
        assert_syntax_error(
            "1.", "Invalid number, expected digit but got: <EOF>.", (1, 3)
        )
        assert_syntax_error(
            "1e", "Invalid number, expected digit but got: <EOF>.", (1, 3)
        )
        assert_syntax_error(
            "1E", "Invalid number, expected digit but got: <EOF>.", (1, 3)
        )
        assert_syntax_error(
            "1.e1", "Invalid number, expected digit but got: 'e'.", (1, 3)
        )
        assert_syntax_error(".123", "Unexpected character: '.'.", (1, 1))
        assert_syntax_error(
            "1.A", "Invalid number, expected digit but got: 'A'.", (1, 3)
        )
        assert_syntax_error(
            "-A", "Invalid number, expected digit but got: 'A'.", (1, 2)
        )
        assert_syntax_error(
            "1.0e", "Invalid number, expected digit but got: <EOF>.", (1, 5)
        )
        assert_syntax_error(
            "1.0eA", "Invalid number, expected digit but got: 'A'.", (1, 5)
        )

        assert_syntax_error(
            '1.0e"', "Invalid number, expected digit but got: '\"'.", (1, 5)
        )

        assert_syntax_error(
            "1.2e3e", "Invalid number, expected digit but got: 'e'.", (1, 6)
        )
        assert_syntax_error(
            "1.2e3.4", "Invalid number, expected digit but got: '.'.", (1, 6)
        )
        assert_syntax_error(
            "1.23.4", "Invalid number, expected digit but got: '.'.", (1, 5)
        )

    def lex_does_not_allow_name_start_after_a_number():
        assert_syntax_error(
            "0xF1", "Invalid number, expected digit but got: 'x'.", (1, 2)
        )
        assert_syntax_error(
            "0b10", "Invalid number, expected digit but got: 'b'.", (1, 2)
        )
        assert_syntax_error(
            "123abc", "Invalid number, expected digit but got: 'a'.", (1, 4)
        )
        assert_syntax_error(
            "1_234", "Invalid number, expected digit but got: '_'.", (1, 2)
        )
        assert_syntax_error("1\xdf", "Unexpected character: U+00DF.", (1, 2))
        assert_syntax_error(
            "1.23f", "Invalid number, expected digit but got: 'f'.", (1, 5)
        )
        assert_syntax_error(
            "1.234_5", "Invalid number, expected digit but got: '_'.", (1, 6)
        )

    # noinspection PyArgumentEqualDefault
    def lexes_punctuation():
        assert lex_one("!") == Token(TokenKind.BANG, 0, 1, 1, 1, None)
        assert lex_one("$") == Token(TokenKind.DOLLAR, 0, 1, 1, 1, None)
        assert lex_one("(") == Token(TokenKind.PAREN_L, 0, 1, 1, 1, None)
        assert lex_one(")") == Token(TokenKind.PAREN_R, 0, 1, 1, 1, None)
        assert lex_one("...") == Token(TokenKind.SPREAD, 0, 3, 1, 1, None)
        assert lex_one(":") == Token(TokenKind.COLON, 0, 1, 1, 1, None)
        assert lex_one("=") == Token(TokenKind.EQUALS, 0, 1, 1, 1, None)
        assert lex_one("@") == Token(TokenKind.AT, 0, 1, 1, 1, None)
        assert lex_one("[") == Token(TokenKind.BRACKET_L, 0, 1, 1, 1, None)
        assert lex_one("]") == Token(TokenKind.BRACKET_R, 0, 1, 1, 1, None)
        assert lex_one("{") == Token(TokenKind.BRACE_L, 0, 1, 1, 1, None)
        assert lex_one("}") == Token(TokenKind.BRACE_R, 0, 1, 1, 1, None)
        assert lex_one("|") == Token(TokenKind.PIPE, 0, 1, 1, 1, None)

    def lex_reports_useful_unknown_character_error():
        assert_syntax_error("..", "Unexpected character: '.'.", (1, 1))
        assert_syntax_error("~", "Unexpected character: '~'.", (1, 1))
        assert_syntax_error("\x00", "Unexpected character: U+0000.", (1, 1))
        assert_syntax_error("\b", "Unexpected character: U+0008.", (1, 1))
        assert_syntax_error("\xAA", "Unexpected character: U+00AA.", (1, 1))
        assert_syntax_error("\u0AAA", "Unexpected character: U+0AAA.", (1, 1))
        assert_syntax_error("\u203B", "Unexpected character: U+203B.", (1, 1))
        assert_syntax_error("\U0001f600", "Unexpected character: U+1F600.", (1, 1))
        assert_syntax_error("\uD83D\uDE00", "Unexpected character: U+1F600.", (1, 1))
        assert_syntax_error("\uD800\uDC00", "Unexpected character: U+10000.", (1, 1))
        assert_syntax_error("\uDBFF\uDFFF", "Unexpected character: U+10FFFF.", (1, 1))
        assert_syntax_error("\uD800", "Invalid character: U+D800.", (1, 1))
        assert_syntax_error("\uDBFF", "Invalid character: U+DBFF.", (1, 1))
        assert_syntax_error("\uDEAD", "Invalid character: U+DEAD.", (1, 1))

    # noinspection PyArgumentEqualDefault
    def lex_reports_useful_information_for_dashes_in_names():
        source = Source("a-b")
        lexer = Lexer(source)
        first_token = lexer.advance()
        assert first_token == Token(TokenKind.NAME, 0, 1, 1, 1, "a")
        with raises(GraphQLSyntaxError) as exc_info:
            lexer.advance()
        error = exc_info.value
        assert error.message == (
            "Syntax Error: Invalid number, expected digit but got: 'b'."
        )
        assert error.locations == [(1, 3)]

    def produces_double_linked_list_of_tokens_including_comments():
        source = Source(
            """
            {
              #comment
              field
            }
            """
        )
        lexer = Lexer(source)
        start_token = lexer.token
        while True:
            end_token = lexer.advance()
            if end_token.kind == TokenKind.EOF:
                break
            assert end_token.kind != TokenKind.COMMENT
        assert start_token.prev is None
        assert end_token.next is None
        tokens: List[Token] = []
        tok: Optional[Token] = start_token
        while tok:
            assert not tokens or tok.prev == tokens[-1]
            tokens.append(tok)
            tok = tok.next
        assert [tok.kind for tok in tokens] == [
            TokenKind.SOF,
            TokenKind.BRACE_L,
            TokenKind.COMMENT,
            TokenKind.NAME,
            TokenKind.BRACE_R,
            TokenKind.EOF,
        ]

    def lexes_comments():
        assert lex_one("# Comment").prev == Token(
            TokenKind.COMMENT, 0, 9, 1, 1, " Comment"
        )
        assert lex_one("# Comment\nAnother line").prev == Token(
            TokenKind.COMMENT, 0, 9, 1, 1, " Comment"
        )
        assert lex_one("# Comment\r\nAnother line").prev == Token(
            TokenKind.COMMENT, 0, 9, 1, 1, " Comment"
        )
        assert lex_one("# Comment \U0001f600").prev == Token(
            TokenKind.COMMENT, 0, 11, 1, 1, " Comment \U0001f600"
        )
        assert lex_one("# Comment \uD83D\uDE00").prev == Token(
            TokenKind.COMMENT, 0, 12, 1, 1, " Comment \uD83D\uDE00"
        )
        assert_syntax_error(
            "# Invalid surrogate \uDEAD", "Invalid character: U+DEAD.", (1, 21)
        )


def describe_is_punctuator_token_kind():
    def _is_punctuator_token(text: str) -> bool:
        return is_punctuator_token_kind(lex_one(text).kind)

    def returns_true_for_punctuator_tokens():
        assert _is_punctuator_token("!") is True
        assert _is_punctuator_token("$") is True
        assert _is_punctuator_token("&") is True
        assert _is_punctuator_token("(") is True
        assert _is_punctuator_token(")") is True
        assert _is_punctuator_token("...") is True
        assert _is_punctuator_token(":") is True
        assert _is_punctuator_token("=") is True
        assert _is_punctuator_token("@") is True
        assert _is_punctuator_token("[") is True
        assert _is_punctuator_token("]") is True
        assert _is_punctuator_token("{") is True
        assert _is_punctuator_token("|") is True
        assert _is_punctuator_token("}") is True

    def returns_false_for_non_punctuator_tokens():
        assert _is_punctuator_token("") is False
        assert _is_punctuator_token("name") is False
        assert _is_punctuator_token("1") is False
        assert _is_punctuator_token("3.14") is False
        assert _is_punctuator_token('"str"') is False
        assert _is_punctuator_token('"""str"""') is False



from functools import partial

from graphql.utilities import build_schema
from graphql.validation import KnownArgumentNamesRule
from graphql.validation.rules.known_argument_names import (
    KnownArgumentNamesOnDirectivesRule,
)

from .harness import assert_validation_errors, assert_sdl_validation_errors

assert_errors = partial(assert_validation_errors, KnownArgumentNamesRule)

assert_valid = partial(assert_errors, errors=[])

assert_sdl_errors = partial(
    assert_sdl_validation_errors, KnownArgumentNamesOnDirectivesRule
)

assert_sdl_valid = partial(assert_sdl_errors, errors=[])


def describe_validate_known_argument_names():
    def single_arg_is_known():
        assert_valid(
            """
            fragment argOnRequiredArg on Dog {
              doesKnowCommand(dogCommand: SIT)
            }
            """
        )

    def multiple_args_are_known():
        assert_valid(
            """
            fragment multipleArgs on ComplicatedArgs {
              multipleReqs(req1: 1, req2: 2)
            }
            """
        )

    def ignore_args_of_unknown_fields():
        assert_valid(
            """
            fragment argOnUnknownField on Dog {
              unknownField(unknownArg: SIT)
            }
            """
        )

    def multiple_args_in_reverse_order_are_known():
        assert_valid(
            """
            fragment multipleArgsReverseOrder on ComplicatedArgs {
              multipleReqs(req2: 2, req1: 1)
            }
            """
        )

    def no_args_on_optional_arg():
        assert_valid(
            """
            fragment noArgOnOptionalArg on Dog {
              isHouseTrained
            }
            """
        )

    def args_are_known_deeply():
        assert_valid(
            """
            {
              dog {
                doesKnowCommand(dogCommand: SIT)
              }
              human {
                pet {
                  ... on Dog {
                      doesKnowCommand(dogCommand: SIT)
                  }
                }
              }
            }
            """
        )

    def directive_args_are_known():
        assert_valid(
            """
            {
              dog @skip(if: true)
            }
            """
        )

    def field_args_are_invalid():
        assert_errors(
            """
            {
              dog @skip(unless: true)
            }
            """,
            [
                {
                    "message": "Unknown argument 'unless' on directive '@skip'.",
                    "locations": [(3, 25)],
                }
            ],
        )

    def directive_without_args_is_valid():
        assert_valid(
            """
            {
                dog @onField
            }
            """
        )

    def arg_passed_to_directive_without_args_is_reported():
        assert_errors(
            """
            {
                dog @onField(if: true)
            }
            """,
            [
                {
                    "message": "Unknown argument 'if' on directive '@onField'.",
                    "locations": [(3, 30)],
                }
            ],
        )

    def misspelled_directive_args_are_reported():
        assert_errors(
            """
            {
              dog @skip(iff: true)
            }
            """,
            [
                {
                    "message": "Unknown argument 'iff' on directive '@skip'."
                    " Did you mean 'if'?",
                    "locations": [(3, 25)],
                }
            ],
        )

    def invalid_arg_name():
        assert_errors(
            """
            fragment invalidArgName on Dog {
              doesKnowCommand(unknown: true)
            }
            """,
            [
                {
                    "message": "Unknown argument 'unknown'"
                    " on field 'Dog.doesKnowCommand'.",
                    "locations": [(3, 31)],
                },
            ],
        )

    def misspelled_arg_name_is_reported():
        assert_errors(
            """
            fragment invalidArgName on Dog {
              doesKnowCommand(DogCommand: true)
            }
            """,
            [
                {
                    "message": "Unknown argument 'DogCommand'"
                    " on field 'Dog.doesKnowCommand'."
                    " Did you mean 'dogCommand'?",
                    "locations": [(3, 31)],
                }
            ],
        )

    def unknown_args_amongst_known_args():
        assert_errors(
            """
            fragment oneGoodArgOneInvalidArg on Dog {
              doesKnowCommand(whoKnows: 1, dogCommand: SIT, unknown: true)
            }
            """,
            [
                {
                    "message": "Unknown argument 'whoKnows'"
                    " on field 'Dog.doesKnowCommand'.",
                    "locations": [(3, 31)],
                },
                {
                    "message": "Unknown argument 'unknown'"
                    " on field 'Dog.doesKnowCommand'.",
                    "locations": [(3, 61)],
                },
            ],
        )

    def unknown_args_deeply():
        assert_errors(
            """
            {
              dog {
                doesKnowCommand(unknown: true)
              }
              human {
                pet {
                  ... on Dog {
                    doesKnowCommand(unknown: true)
                  }
                }
              }
            }
            """,
            [
                {
                    "message": "Unknown argument 'unknown'"
                    " on field 'Dog.doesKnowCommand'.",
                    "locations": [(4, 33)],
                },
                {
                    "message": "Unknown argument 'unknown'"
                    " on field 'Dog.doesKnowCommand'.",
                    "locations": [(9, 37)],
                },
            ],
        )

    def describe_within_sdl():
        def known_arg_on_directive_inside_sdl():
            assert_sdl_valid(
                """
                type Query {
                  foo: String @test(arg: "")
                }

                directive @test(arg: String) on FIELD_DEFINITION
                """
            )

        def unknown_arg_on_directive_defined_inside_sdl():
            assert_sdl_errors(
                """
                type Query {
                  foo: String @test(unknown: "")
                }

                directive @test(arg: String) on FIELD_DEFINITION
                """,
                [
                    {
                        "message": "Unknown argument 'unknown' on directive '@test'.",
                        "locations": [(3, 37)],
                    },
                ],
            )

        def misspelled_arg_name_is_reported_on_directive_defined_inside_sdl():
            assert_sdl_errors(
                """
                type Query {
                  foo: String @test(agr: "")
                }

                directive @test(arg: String) on FIELD_DEFINITION
                """,
                [
                    {
                        "message": "Unknown argument 'agr' on directive '@test'."
                        " Did you mean 'arg'?",
                        "locations": [(3, 37)],
                    },
                ],
            )

        def unknown_arg_on_standard_directive():
            assert_sdl_errors(
                """
                type Query {
                  foo: String @deprecated(unknown: "")
                }
                """,
                [
                    {
                        "message": "Unknown argument 'unknown'"
                        " on directive '@deprecated'.",
                        "locations": [(3, 43)],
                    },
                ],
            )

        def unknown_arg_on_overridden_standard_directive():
            assert_sdl_errors(
                """
                type Query {
                  foo: String @deprecated(reason: "")
                }
                directive @deprecated(arg: String) on FIELD
                """,
                [
                    {
                        "message": "Unknown argument 'reason'"
                        " on directive '@deprecated'.",
                        "locations": [(3, 43)],
                    },
                ],
            )

        def unknown_arg_on_directive_defined_in_schema_extension():
            schema = build_schema(
                """
                type Query {
                  foo: String
                }
                """
            )
            assert_sdl_errors(
                """
                directive @test(arg: String) on OBJECT

                extend type Query  @test(unknown: "")
                """,
                [
                    {
                        "message": "Unknown argument 'unknown' on directive '@test'.",
                        "locations": [(4, 42)],
                    },
                ],
                schema,
            )

        def unknown_arg_on_directive_used_in_schema_extension():
            schema = build_schema(
                """
                directive @test(arg: String) on OBJECT

                type Query {
                  foo: String
                }
                """
            )
            assert_sdl_errors(
                """
                extend type Query @test(unknown: "")
                """,
                [
                    {
                        "message": "Unknown argument 'unknown' on directive '@test'.",
                        "locations": [(2, 41)],
                    },
                ],
                schema,
            )



from collections import defaultdict
from typing import cast, Any, Dict

from ...error import GraphQLError
from ...language import NameNode, EnumTypeDefinitionNode, VisitorAction, SKIP
from ...type import is_enum_type, GraphQLEnumType
from . import SDLValidationContext, SDLValidationRule

__all__ = ["UniqueEnumValueNamesRule"]


class UniqueEnumValueNamesRule(SDLValidationRule):
    """Unique enum value names

    A GraphQL enum type is only valid if all its values are uniquely named.
    """

    def __init__(self, context: SDLValidationContext):
        super().__init__(context)
        schema = context.schema
        self.existing_type_map = schema.type_map if schema else {}
        self.known_value_names: Dict[str, Dict[str, NameNode]] = defaultdict(dict)

    def check_value_uniqueness(
        self, node: EnumTypeDefinitionNode, *_args: Any
    ) -> VisitorAction:
        existing_type_map = self.existing_type_map
        type_name = node.name.value
        value_names = self.known_value_names[type_name]

        for value_def in node.values or []:
            value_name = value_def.name.value

            existing_type = existing_type_map.get(type_name)
            if (
                is_enum_type(existing_type)
                and value_name in cast(GraphQLEnumType, existing_type).values
            ):
                self.report_error(
                    GraphQLError(
                        f"Enum value '{type_name}.{value_name}'"
                        " already exists in the schema."
                        " It cannot also be defined in this type extension.",
                        value_def.name,
                    )
                )
            elif value_name in value_names:
                self.report_error(
                    GraphQLError(
                        f"Enum value '{type_name}.{value_name}'"
                        " can only be defined once.",
                        [value_names[value_name], value_def.name],
                    )
                )
            else:
                value_names[value_name] = value_def.name

        return SKIP

    enter_enum_type_definition = check_value_uniqueness
    enter_enum_type_extension = check_value_uniqueness



from graphql.execution import execute_sync, ExecutionResult
from graphql.language import parse
from graphql.type import GraphQLObjectType, GraphQLField, GraphQLSchema, GraphQLString

schema = GraphQLSchema(
    GraphQLObjectType(
        "TestType", {"a": GraphQLField(GraphQLString), "b": GraphQLField(GraphQLString)}
    )
)


# noinspection PyMethodMayBeStatic
class RootValue:
    def a(self, *_args):
        return "a"

    def b(self, *_args):
        return "b"


def execute_test_query(query: str) -> ExecutionResult:
    document = parse(query)
    return execute_sync(schema, document, RootValue())


def describe_execute_handles_directives():
    def describe_works_without_directives():
        def basic_query_works():
            result = execute_test_query("{ a, b }")
            assert result == ({"a": "a", "b": "b"}, None)

    def describe_works_on_scalars():
        def if_true_includes_scalar():
            result = execute_test_query("{ a, b @include(if: true) }")
            assert result == ({"a": "a", "b": "b"}, None)

        def if_false_omits_on_scalar():
            result = execute_test_query("{ a, b @include(if: false) }")
            assert result == ({"a": "a"}, None)

        def unless_false_includes_scalar():
            result = execute_test_query("{ a, b @skip(if: false) }")
            assert result == ({"a": "a", "b": "b"}, None)

        def unless_true_omits_scalar():
            result = execute_test_query("{ a, b @skip(if: true) }")
            assert result == ({"a": "a"}, None)

    def describe_works_on_fragment_spreads():
        def if_false_omits_fragment_spread():
            result = execute_test_query(
                """
                query Q {
                  a
                  ...Frag @include(if: false)
                }
                fragment Frag on TestType {
                  b
                }
                """
            )
            assert result == ({"a": "a"}, None)

        def if_true_includes_fragment_spread():
            result = execute_test_query(
                """
                query Q {
                  a
                  ...Frag @include(if: true)
                }
                fragment Frag on TestType {
                  b
                }
                """
            )
            assert result == ({"a": "a", "b": "b"}, None)

        def unless_false_includes_fragment_spread():
            result = execute_test_query(
                """
                query Q {
                  a
                  ...Frag @skip(if: false)
                }
                fragment Frag on TestType {
                  b
                }
                """
            )
            assert result == ({"a": "a", "b": "b"}, None)

        def unless_true_omits_fragment_spread():
            result = execute_test_query(
                """
                query Q {
                  a
                  ...Frag @skip(if: true)
                }
                fragment Frag on TestType {
                  b
                }
                """
            )
            assert result == ({"a": "a"}, None)

    def describe_works_on_inline_fragment():
        def if_false_omits_inline_fragment():
            result = execute_test_query(
                """
                query Q {
                  a
                  ... on TestType @include(if: false) {
                    b
                  }
                }
                """
            )
            assert result == ({"a": "a"}, None)

        def if_true_includes_inline_fragment():
            result = execute_test_query(
                """
                query Q {
                  a
                  ... on TestType @include(if: true) {
                    b
                  }
                }
                """
            )
            assert result == ({"a": "a", "b": "b"}, None)

        def unless_false_includes_inline_fragment():
            result = execute_test_query(
                """
                query Q {
                  a
                  ... on TestType @skip(if: false) {
                    b
                  }
                }
                """
            )
            assert result == ({"a": "a", "b": "b"}, None)

        def unless_true_omits_inline_fragment():
            result = execute_test_query(
                """
                query Q {
                  a
                  ... on TestType @skip(if: true) {
                    b
                  }
                }
                """
            )
            assert result == ({"a": "a"}, None)

    def describe_works_on_anonymous_inline_fragment():
        def if_false_omits_anonymous_inline_fragment():
            result = execute_test_query(
                """
                query {
                  a
                  ... @include(if: false) {
                    b
                  }
                }
                """
            )
            assert result == ({"a": "a"}, None)

        def if_true_includes_anonymous_inline_fragment():
            result = execute_test_query(
                """
                query {
                  a
                  ... @include(if: true) {
                    b
                  }
                }
                """
            )
            assert result == ({"a": "a", "b": "b"}, None)

        def unless_false_includes_anonymous_inline_fragment():
            result = execute_test_query(
                """
                query {
                  a
                  ... @skip(if: false) {
                    b
                  }
                }
                """
            )
            assert result == ({"a": "a", "b": "b"}, None)

        def unless_true_omits_anonymous_inline_fragment():
            result = execute_test_query(
                """
                query {
                  a
                  ... @skip(if: true) {
                    b
                  }
                }
                """
            )
            assert result == ({"a": "a"}, None)

    def describe_works_with_skip_and_include_directives():
        def include_and_no_skip():
            result = execute_test_query(
                """
                {
                  a
                  b @include(if: true) @skip(if: false)
                }
                """
            )
            assert result == ({"a": "a", "b": "b"}, None)

        def include_and_skip():
            result = execute_test_query(
                """
                {
                  a
                  b @include(if: true) @skip(if: true)
                }
                """
            )
            assert result == ({"a": "a"}, None)

        def no_include_or_skip():
            result = execute_test_query(
                """
                {
                  a
                  b @include(if: false) @skip(if: false)
                }
                """
            )
            assert result == ({"a": "a"}, None)



from typing import Optional, Sequence

__all__ = ["did_you_mean"]

MAX_LENGTH = 5


def did_you_mean(suggestions: Sequence[str], sub_message: Optional[str] = None) -> str:
    """Given [ A, B, C ] return ' Did you mean A, B, or C?'"""
    if not suggestions or not MAX_LENGTH:
        return ""
    parts = [" Did you mean "]
    if sub_message:
        parts.extend([sub_message, " "])
    suggestions = suggestions[:MAX_LENGTH]
    n = len(suggestions)
    if n == 1:
        parts.append(f"'{suggestions[0]}'?")
    elif n == 2:
        parts.append(f"'{suggestions[0]}' or '{suggestions[1]}'?")
    else:
        parts.extend(
            [
                ", ".join(f"'{s}'" for s in suggestions[:-1]),
                f", or '{suggestions[-1]}'?",
            ]
        )
    return "".join(parts)



from graphql.pyutils import merge_kwargs

try:
    from typing import TypedDict
except ImportError:  # Python < 3.8
    from typing_extensions import TypedDict


class FooDict(TypedDict):
    foo: str
    bar: str
    baz: int


def describe_merge_kwargs():
    def should_merge_with_no_kwargs():
        base = FooDict(foo="foo", bar="bar", baz=0)
        assert merge_kwargs(base) == base

    def should_merge_with_kwargs():
        base = FooDict(foo="foo", bar="bar", baz=0)
        assert merge_kwargs(base, foo="moo", bar="mar", baz=1) == FooDict(
            foo="moo", bar="mar", baz=1
        )



from typing import List

from graphql.pyutils import suggestion_list


def expect_suggestions(input_: str, options: List[str], expected: List[str]) -> None:
    assert suggestion_list(input_, options) == expected


def describe_suggestion_list():
    def returns_results_when_input_is_empty():
        expect_suggestions("", ["a"], ["a"])

    def returns_empty_array_when_there_are_no_options():
        expect_suggestions("input", [], [])

    def returns_options_with_small_lexical_distance():
        expect_suggestions("greenish", ["green"], ["green"])
        expect_suggestions("green", ["greenish"], ["greenish"])

    def rejects_options_with_distance_that_exceeds_threshold():
        expect_suggestions("aaaa", ["aaab"], ["aaab"])
        expect_suggestions("aaaa", ["aabb"], ["aabb"])
        expect_suggestions("aaaa", ["abbb"], [])

        expect_suggestions("ab", ["ca"], [])

    def returns_options_with_different_case():
        expect_suggestions("verylongstring", ["VERYLONGSTRING"], ["VERYLONGSTRING"])

        expect_suggestions("VERYLONGSTRING", ["verylongstring"], ["verylongstring"])

        expect_suggestions("VERYLONGSTRING", ["VeryLongString"], ["VeryLongString"])

    def returns_options_with_transpositions():
        expect_suggestions("agr", ["arg"], ["arg"])

        expect_suggestions("214365879", ["123456789"], ["123456789"])

    def returns_options_sorted_based_on_lexical_distance():
        expect_suggestions("abc", ["a", "ab", "abc"], ["abc", "ab", "a"])

        expect_suggestions(
            "GraphQl",
            ["graphics", "SQL", "GraphQL", "quarks", "mark"],
            ["GraphQL", "graphics"],
        )

    def returns_options_with_the_same_lexical_distance_sorted_naturally():
        expect_suggestions("a", ["az", "ax", "ay"], ["ax", "ay", "az"])

        expect_suggestions("boo", ["moo", "foo", "zoo"], ["foo", "moo", "zoo"])

        expect_suggestions("abc", ["a1", "a12", "a2"], ["a1", "a2", "a12"])

    def returns_options_sorted_first_by_lexical_distance_then_naturally():
        expect_suggestions(
            "csutomer",
            ["store", "customer", "stomer", "some", "more"],
            ["customer", "stomer", "some", "store"],
        )



from functools import partial

from graphql.language import parse
from graphql.utilities import extend_schema
from graphql.validation import UniqueDirectivesPerLocationRule

from .harness import assert_validation_errors, assert_sdl_validation_errors, test_schema

extension_sdl = """
  directive @directive on FIELD | FRAGMENT_DEFINITION
  directive @directiveA on FIELD | FRAGMENT_DEFINITION
  directive @directiveB on FIELD | FRAGMENT_DEFINITION
  directive @repeatable repeatable on FIELD | FRAGMENT_DEFINITION
"""
schema_with_directives = extend_schema(test_schema, parse(extension_sdl))

assert_errors = partial(
    assert_validation_errors,
    UniqueDirectivesPerLocationRule,
    schema=schema_with_directives,
)

assert_valid = partial(assert_errors, errors=[])

assert_sdl_errors = partial(
    assert_sdl_validation_errors, UniqueDirectivesPerLocationRule
)


def describe_validate_directives_are_unique_per_location():
    def no_directives():
        assert_valid(
            """
            {
              field
            }
            """
        )

    def unique_directives_in_different_locations():
        assert_valid(
            """
            fragment Test on Type @directiveA {
              field @directiveB
            }
            """
        )

    def unique_directives_in_same_locations():
        assert_valid(
            """
            fragment Test on Type @directiveA @directiveB {
              field @directiveA @directiveB
            }
            """
        )

    def same_directives_in_different_locations():
        assert_valid(
            """
            fragment Test on Type @directiveA {
              field @directiveA
            }
            """
        )

    def same_directives_in_similar_locations():
        assert_valid(
            """
            fragment Test on Type {
              field @directive
              field @directive
            }
            """
        )

    def repeatable_directives_in_same_location():
        assert_valid(
            """
            fragment Test on Type @repeatable @repeatable {
              field @repeatable @repeatable
            }
            """
        )

    def unknown_directives_must_be_ignored():
        assert_valid(
            """
            type Test @unknown @unknown {
              field: String! @unknown @unknown
            }

            extend type Test @unknown {
              anotherField: String!
            }
            """
        )

    def duplicate_directives_in_one_location():
        assert_errors(
            """
            fragment Test on Type {
              field @directive @directive
            }
            """,
            [
                {
                    "message": "The directive '@directive'"
                    " can only be used once at this location.",
                    "locations": [(3, 21), (3, 32)],
                },
            ],
        )

    def many_duplicate_directives_in_one_location():
        assert_errors(
            """
            fragment Test on Type {
              field @directive @directive @directive
            }
            """,
            [
                {
                    "message": "The directive '@directive'"
                    " can only be used once at this location.",
                    "locations": [(3, 21), (3, 32)],
                },
                {
                    "message": "The directive '@directive'"
                    " can only be used once at this location.",
                    "locations": [(3, 21), (3, 43)],
                },
            ],
        )

    def different_duplicate_directives_in_one_location():
        assert_errors(
            """
            fragment Test on Type {
              field @directiveA @directiveB @directiveA @directiveB
            }
            """,
            [
                {
                    "message": "The directive '@directiveA'"
                    " can only be used once at this location.",
                    "locations": [(3, 21), (3, 45)],
                },
                {
                    "message": "The directive '@directiveB'"
                    " can only be used once at this location.",
                    "locations": [(3, 33), (3, 57)],
                },
            ],
        )

    def different_duplicate_directives_in_many_locations():
        assert_errors(
            """
            fragment Test on Type @directive @directive {
              field @directive @directive
            }
            """,
            [
                {
                    "message": "The directive '@directive'"
                    " can only be used once at this location.",
                    "locations": [(2, 35), (2, 46)],
                },
                {
                    "message": "The directive '@directive'"
                    " can only be used once at this location.",
                    "locations": [(3, 21), (3, 32)],
                },
            ],
        )

    def duplicate_directives_on_sdl_definitions():
        assert_sdl_errors(
            """
            directive @nonRepeatable on
              SCHEMA | SCALAR | OBJECT | INTERFACE | UNION | INPUT_OBJECT

            schema @nonRepeatable @nonRepeatable { query: Dummy }

            scalar TestScalar @nonRepeatable @nonRepeatable
            type TestObject @nonRepeatable @nonRepeatable
            interface TestInterface @nonRepeatable @nonRepeatable
            union TestUnion @nonRepeatable @nonRepeatable
            input TestInput @nonRepeatable @nonRepeatable
            """,
            [
                {
                    "message": "The directive '@nonRepeatable'"
                    " can only be used once at this location.",
                    "locations": [(5, 20), (5, 35)],
                },
                {
                    "message": "The directive '@nonRepeatable'"
                    " can only be used once at this location.",
                    "locations": [(7, 31), (7, 46)],
                },
                {
                    "message": "The directive '@nonRepeatable'"
                    " can only be used once at this location.",
                    "locations": [(8, 29), (8, 44)],
                },
                {
                    "message": "The directive '@nonRepeatable'"
                    " can only be used once at this location.",
                    "locations": [(9, 37), (9, 52)],
                },
                {
                    "message": "The directive '@nonRepeatable'"
                    " can only be used once at this location.",
                    "locations": [(10, 29), (10, 44)],
                },
                {
                    "message": "The directive '@nonRepeatable'"
                    " can only be used once at this location.",
                    "locations": [(11, 29), (11, 44)],
                },
            ],
        )

    def duplicate_directives_on_sdl_extensions():
        assert_sdl_errors(
            """
            directive @nonRepeatable on
              SCHEMA | SCALAR | OBJECT | INTERFACE | UNION | INPUT_OBJECT

            extend schema @nonRepeatable @nonRepeatable

            extend scalar TestScalar @nonRepeatable @nonRepeatable
            extend type TestObject @nonRepeatable @nonRepeatable
            extend interface TestInterface @nonRepeatable @nonRepeatable
            extend union TestUnion @nonRepeatable @nonRepeatable
            extend input TestInput @nonRepeatable @nonRepeatable
            """,
            [
                {
                    "message": "The directive '@nonRepeatable'"
                    " can only be used once at this location.",
                    "locations": [(5, 27), (5, 42)],
                },
                {
                    "message": "The directive '@nonRepeatable'"
                    " can only be used once at this location.",
                    "locations": [(7, 38), (7, 53)],
                },
                {
                    "message": "The directive '@nonRepeatable'"
                    " can only be used once at this location.",
                    "locations": [(8, 36), (8, 51)],
                },
                {
                    "message": "The directive '@nonRepeatable'"
                    " can only be used once at this location.",
                    "locations": [(9, 44), (9, 59)],
                },
                {
                    "message": "The directive '@nonRepeatable'"
                    " can only be used once at this location.",
                    "locations": [(10, 36), (10, 51)],
                },
                {
                    "message": "The directive '@nonRepeatable'"
                    " can only be used once at this location.",
                    "locations": [(11, 36), (11, 51)],
                },
            ],
        )

    def duplicate_directives_between_sdl_definitions_and_extensions():
        assert_sdl_errors(
            """
            directive @nonRepeatable on SCHEMA

            schema @nonRepeatable { query: Dummy }
            extend schema @nonRepeatable
            """,
            [
                {
                    "message": "The directive '@nonRepeatable'"
                    " can only be used once at this location.",
                    "locations": [(4, 20), (5, 27)],
                },
            ],
        )

        assert_sdl_errors(
            """
            directive @nonRepeatable on SCALAR

            scalar TestScalar @nonRepeatable
            extend scalar TestScalar @nonRepeatable
            scalar TestScalar @nonRepeatable
            """,
            [
                {
                    "message": "The directive '@nonRepeatable'"
                    " can only be used once at this location.",
                    "locations": [(4, 31), (5, 38)],
                },
                {
                    "message": "The directive '@nonRepeatable'"
                    " can only be used once at this location.",
                    "locations": [(4, 31), (6, 31)],
                },
            ],
        )

        assert_sdl_errors(
            """
            directive @nonRepeatable on OBJECT

            extend type TestObject @nonRepeatable
            type TestObject @nonRepeatable
            extend type TestObject @nonRepeatable
            """,
            [
                {
                    "message": "The directive '@nonRepeatable'"
                    " can only be used once at this location.",
                    "locations": [(4, 36), (5, 29)],
                },
                {
                    "message": "The directive '@nonRepeatable'"
                    " can only be used once at this location.",
                    "locations": [(4, 36), (6, 36)],
                },
            ],
        )



from pytest import raises

from graphql.error import GraphQLError
from graphql.language import (
    parse,
    DocumentNode,
    OperationDefinitionNode,
    OperationTypeDefinitionNode,
    SchemaDefinitionNode,
)
from graphql.type import GraphQLField, GraphQLObjectType, GraphQLSchema, GraphQLString
from graphql.utilities import get_operation_root_type


query_type = GraphQLObjectType("FooQuery", {"field": GraphQLField(GraphQLString)})

mutation_type = GraphQLObjectType("FooMutation", {"field": GraphQLField(GraphQLString)})

subscription_type = GraphQLObjectType(
    "FooSubscription", {"field": GraphQLField(GraphQLString)}
)


def get_operation_node(doc: DocumentNode) -> OperationDefinitionNode:
    operation_node = doc.definitions[0]
    assert isinstance(operation_node, OperationDefinitionNode)
    return operation_node


def describe_deprecated_get_operation_root_type():
    def gets_a_query_type_for_an_unnamed_operation_definition_node():
        test_schema = GraphQLSchema(query_type)
        doc = parse("{ field }")
        operation_node = get_operation_node(doc)
        assert get_operation_root_type(test_schema, operation_node) is query_type

    def gets_a_query_type_for_a_named_operation_definition_node():
        test_schema = GraphQLSchema(query_type)
        doc = parse("query Q { field }")
        operation_node = get_operation_node(doc)
        assert get_operation_root_type(test_schema, operation_node) is query_type

    def gets_a_type_for_operation_definition_nodes():
        test_schema = GraphQLSchema(query_type, mutation_type, subscription_type)
        doc = parse(
            """
            schema {
              query: FooQuery
              mutation: FooMutation
              subscription: FooSubscription
            }
            """
        )

        schema_node = doc.definitions[0]
        assert isinstance(schema_node, SchemaDefinitionNode)
        query_node, mutation_node, subscription_node = schema_node.operation_types
        assert isinstance(query_node, OperationTypeDefinitionNode)
        assert get_operation_root_type(test_schema, query_node) is query_type
        assert isinstance(mutation_node, OperationTypeDefinitionNode)
        assert get_operation_root_type(test_schema, mutation_node) is mutation_type
        assert isinstance(subscription_node, OperationTypeDefinitionNode)
        assert (
            get_operation_root_type(test_schema, subscription_node) is subscription_type
        )

    def gets_a_mutation_type_for_an_operation_definition_node():
        test_schema = GraphQLSchema(mutation=mutation_type)
        doc = parse("mutation { field }")
        operation_node = get_operation_node(doc)
        assert get_operation_root_type(test_schema, operation_node) is mutation_type

    def gets_a_subscription_type_for_an_operation_definition_node():
        test_schema = GraphQLSchema(subscription=subscription_type)
        doc = parse("subscription { field }")
        operation_node = get_operation_node(doc)
        assert get_operation_root_type(test_schema, operation_node) is subscription_type

    def throws_when_query_type_not_defined_in_schema():
        test_schema = GraphQLSchema()
        doc = parse("query { field }")
        operation_node = get_operation_node(doc)
        with raises(GraphQLError) as exc_info:
            get_operation_root_type(test_schema, operation_node)
        assert exc_info.value.message == (
            "Schema does not define the required query root type."
        )

    def throws_when_mutation_type_not_defined_in_schema():
        test_schema = GraphQLSchema()
        doc = parse("mutation { field }")
        operation_node = get_operation_node(doc)
        with raises(GraphQLError) as exc_info:
            get_operation_root_type(test_schema, operation_node)
        assert exc_info.value.message == "Schema is not configured for mutations."

    def throws_when_subscription_type_not_defined_in_schema():
        test_schema = GraphQLSchema()
        doc = parse("subscription { field }")
        operation_node = get_operation_node(doc)
        with raises(GraphQLError) as exc_info:
            get_operation_root_type(test_schema, operation_node)
        assert exc_info.value.message == "Schema is not configured for subscriptions."

    def throws_when_operation_not_a_valid_operation_kind():
        test_schema = GraphQLSchema()
        doc = parse("{ field }")
        operation_node = get_operation_node(doc)
        operation_node.operation = "non_existent_operation"  # type: ignore
        with raises(GraphQLError) as exc_info:
            get_operation_root_type(test_schema, operation_node)
        assert exc_info.value.message == (
            "Can only have query, mutation and subscription operations."
        )



from graphql import parse, build_ast_schema, GraphQLSchema

from ..fixtures import big_schema_sdl  # noqa: F401


def test_build_schema_from_ast(benchmark, big_schema_sdl):  # noqa: F811
    schema_ast = parse(big_schema_sdl)
    schema: GraphQLSchema = benchmark(
        lambda: build_ast_schema(schema_ast, assume_valid=True)
    )
    assert schema.query_type is not None



import re
from typing import Tuple

from itertools import cycle

__all__ = ["natural_comparison_key"]

_re_digits = re.compile(r"(\d+)")


def natural_comparison_key(key: str) -> Tuple:
    """Comparison key function for sorting strings by natural sort order.

    See: https://en.wikipedia.org/wiki/Natural_sort_order
    """
    return tuple(
        (int(part), part) if is_digit else part
        for part, is_digit in zip(_re_digits.split(key), cycle((False, True)))
    )



from pytest import raises

from graphql.error import GraphQLError
from graphql.language import parse
from graphql.utilities import TypeInfo, build_schema
from graphql.validation import ValidationRule, validate

from .harness import test_schema


def describe_validate_supports_full_validation():
    def rejects_invalid_documents():
        with raises(TypeError) as exc_info:
            # noinspection PyTypeChecker
            assert validate(test_schema, None)  # type: ignore
        assert str(exc_info.value) == "Must provide document."

    def rejects_invalid_type_info():
        with raises(TypeError) as exc_info:
            # noinspection PyTypeChecker
            assert validate(
                test_schema, parse("query { name }"), type_info={}  # type: ignore
            )
        assert str(exc_info.value) == "Not a TypeInfo object: {}."

    def rejects_invalid_rules():
        with raises(TypeError) as exc_info:
            # noinspection PyTypeChecker
            assert validate(
                test_schema, parse("query { name }"), rules=[None]  # type: ignore
            )
        assert (
            str(exc_info.value) == "Rules must be specified as a collection"
            " of ASTValidationRule subclasses."
        )

    def rejects_invalid_max_errors():
        with raises(TypeError) as exc_info:
            # noinspection PyTypeChecker
            assert validate(
                test_schema, parse("query { name }"), max_errors=2.5  # type: ignore
            )
        assert (
            str(exc_info.value)
            == "The maximum number of errors must be passed as an int."
        )

    def validates_queries():
        doc = parse(
            """
            query {
              human {
                pets {
                  ... on Cat {
                    meowsVolume
                  }
                  ... on Dog {
                    barkVolume
                  }
                }
              }
            }
            """
        )

        errors = validate(test_schema, doc)
        assert errors == []

    def detects_unknown_fields():
        doc = parse(
            """
            {
              unknown
            }
            """
        )

        errors = validate(test_schema, doc)
        assert errors == [
            {"message": "Cannot query field 'unknown' on type 'QueryRoot'."}
        ]

    def deprecated_validates_using_a_custom_type_info():
        # This TypeInfo will never return a valid field.
        type_info = TypeInfo(test_schema, None, lambda *args: None)

        doc = parse(
            """
            query {
              human {
                pets {
                  ... on Cat {
                    meowsVolume
                  }
                  ... on Dog {
                    barkVolume
                  }
                }
              }
            }
            """
        )

        errors = validate(test_schema, doc, None, None, type_info)

        assert [error.message for error in errors] == [
            "Cannot query field 'human' on type 'QueryRoot'. Did you mean 'human'?",
            "Cannot query field 'meowsVolume' on type 'Cat'."
            " Did you mean 'meowsVolume'?",
            "Cannot query field 'barkVolume' on type 'Dog'."
            " Did you mean 'barkVolume'?",
        ]

    def validates_using_a_custom_rule():
        schema = build_schema(
            """
            directive @custom(arg: String) on FIELD

            type Query {
              foo: String
            }
            """
        )

        doc = parse(
            """
            query {
              name @custom
            }
            """
        )

        class CustomRule(ValidationRule):
            def enter_directive(self, node, *_args):
                directive_def = self.context.get_directive()
                error = GraphQLError("Reporting directive: " + str(directive_def), node)
                self.context.report_error(error)

        errors = validate(schema, doc, [CustomRule])
        assert errors == [
            {"message": "Reporting directive: @custom", "locations": [(3, 20)]}
        ]


def describe_validate_limit_maximum_number_of_validation_errors():
    query = """
        {
          firstUnknownField
          secondUnknownField
          thirdUnknownField
        }
        """
    doc = parse(query, no_location=True)

    def _validate_document(max_errors=None):
        return validate(test_schema, doc, max_errors=max_errors)

    def _invalid_field_error(field_name: str):
        return {
            "message": f"Cannot query field '{field_name}' on type 'QueryRoot'.",
        }

    def when_max_errors_is_equal_to_number_of_errors():
        errors = _validate_document(max_errors=3)
        assert errors == [
            _invalid_field_error("firstUnknownField"),
            _invalid_field_error("secondUnknownField"),
            _invalid_field_error("thirdUnknownField"),
        ]

    def when_max_errors_is_less_than_number_of_errors():
        errors = _validate_document(max_errors=2)
        assert errors == [
            _invalid_field_error("firstUnknownField"),
            _invalid_field_error("secondUnknownField"),
            {
                "message": "Too many validation errors, error limit reached."
                " Validation aborted."
            },
        ]

    def limits_to_100_when_max_errors_is_not_passed():
        errors = validate(
            test_schema,
            parse(
                "{" + " ".join(f"unknownField{n}" for n in range(120)) + "}",
                no_location=True,
            ),
        )
        assert len(errors) == 101
        assert errors[0] == _invalid_field_error("unknownField0")
        assert errors[-2] == _invalid_field_error("unknownField99")
        assert errors[-1] == {
            "message": "Too many validation errors, error limit reached."
            " Validation aborted."
        }

    def pass_through_exceptions_from_rules():
        class CustomRule(ValidationRule):
            def enter_field(self, *_args):
                raise RuntimeError("Error from custom rule!")

        with raises(RuntimeError, match="^Error from custom rule!$"):
            validate(test_schema, doc, [CustomRule], max_errors=1)
